/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include <iostream>

#include "ipanemaSolvers/ipanemaSolver.h"
#include "ipanemaObjectives/integralOfControl.h"
#include "ipanemaConstraints/samplingProbabilityConstraint.h"
#include "PropagatorSurrogates/intrusivePolynomialPropagator.h"

#include "Dynamics/base_dynamics.h"
#include "Integrators/rk4.h"
#include "Utils/mixed_functions.h"

#include "Astro-Core/conversion_coordinates.h"

#include "OrbitObservations/orbitDistribution.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

using namespace std;
using namespace ipanema;
using namespace propagatorSurrogates;
using namespace ipanemaConstraints;
using namespace ipanemaObjectives;
using namespace ipanemaSolvers;
using namespace utopiaeorbit::orbitobservations;




/**
 * dynamics in equinoctial elements
 */
enum fableDynamicsType {
    inTime,
    inLongitude,
    maxType
};



static const double au = 149597870.691*1.0e3;    // [m]
static const double Tsid = 86164.090517;       // [s]
static const double mu_sun = 1.32712440018e20; // [m3/s2]

/**
 * Dimension of the problem
 */

static unsigned int nc       = 3;             // Number of control variables
static unsigned int ng       = 1;             // Number of constraints
static unsigned int nSamples = 5000;          // Number of samples
static          int nDeg     = 4;
static         bool monomial = true;
static       double intStep  = 2.0*60.0*60.0; // 2 hour integration step
static          int nint     = (int)ceil((54259200.003096-0.0)/intStep);   // Number of integration steps for the whole interval

// Scales
//static double L_scale = 1.0;
//static double T_scale = 1.0;
static double L_scale = au;
static double T_scale = 1.0e8;

/**
 * Fixed quantity
 */
static double maxT       = 0.053 / L_scale * pow(T_scale,2.0) ; // Maximum thrust FABLE [N=kg*m/s^2]
static double m0         = 644.3; // Initial mass [kg]
//static unsigned int nRev = 1;
static double mu         = mu_sun / pow(L_scale,3.0) * pow(T_scale,2.0); // Sun gravity constant [m^3/(s^2)]

static double maxDR      = 3.0e8/L_scale;
static double maxDV      = 1.0e2/L_scale*T_scale;

static bool   finalConditionInEq = false;

/**
 * Reference V infinity condition
 */
static double     v_inf = +3.34e3; // [m/s]
static double alpha_inf = +0.670034153413783;      // [rad]
static double  beta_inf = -0.500473251647287;      // [rad]
//static double     v_infScaled = v_inf/L_scale*T_scale; // [m/s]

// V infinity deviations
static double     v_infMinDev = -100.0;          // [m/s]
static double alpha_infMinDev = +1.0*M_PI/180.0; // [rad]
static double  beta_infMinDev = +0.5*M_PI/180.0; // [rad]

// Type of dynamics to use
static fableDynamicsType dynamicsType_ = inLongitude;




template <class StateType>
class fableDynamics : public smartmath::dynamics::base_dynamics<StateType> {

public:


    struct fableDynamicsParams {

        // Gravitational parametr
        double                           mu;

        // Time grid
        vector<double>*                  pSGrid;

        // Inner vector gathers for control coefficients per each variable
        vector<vector<StateType>>*       pControlCoefficients;

    };


protected:

    const fableDynamicsParams* m_pParams;

    fableDynamicsType          m_type;

public:

    explicit fableDynamics(const fableDynamicsParams* pParams,
                           const fableDynamicsType &type = inLongitude ) :
            smartmath::dynamics::base_dynamics<StateType>("fableDynamics"),
            m_pParams(pParams),
            m_type(type)
    {

    }

    ~fableDynamics() = default;

    /**
     * Get independent variable s index
     */
    unsigned int get_sIndex(const double &s) const
    {

        unsigned int ni = (*m_pParams->pSGrid).size()-1;

        double s0, sf;
        for ( auto i = 0 ; i < ni ; i++ )
        {
            s0 = (*m_pParams->pSGrid)[i];
            sf = (*m_pParams->pSGrid)[i+1];
            if ( s >= s0 && s < sf )
                return i;
        }

        if ( s >= s0 && (s-sf)/(std::min(std::abs(s),std::abs(sf))<1e-10 ))
            return ni-1;

        ipanema_throw("Input independent variable is outside defined grid: " + "input s = " + std::to_string(s) +
                      "; S grid = [" +
                      std::to_string((*m_pParams->pSGrid)[0]) +
                      "," + std::to_string((*m_pParams->pSGrid)[ni]) + "]" );
        return 10000;
    }


    /**
     * @brief evaluate evaluate the dynamics at a given independent variable s and a given state.
     *
     * Function to evaluate the dinamics at a given independent variable s and a given state.
     * @param[in] s Independent variable
     * @param[in] state state values at independent variable s
     * @param[out] dstate derivative of the states at independent variable s
     * @return
     */
    int evaluate(const double &s, const vector<StateType> &x, vector<StateType> &dx) const override {

        auto nc = (*m_pParams->pControlCoefficients)[0].size();
        // Sanity check
        if(nc!=3)
            ipanema_throw("3 control variables are mandatory");

        // Time index which indicates discretisation interval
        auto sIndex = get_sIndex(s);

        // Control vector [magnitude-angle]
        vector<StateType> u((*m_pParams->pControlCoefficients)[sIndex]) ;

        // Transformed control vector [components]
        vector<StateType> U ;
        U.push_back( u[0]*cos(u[1])*cos(u[2]) );
        U.push_back( u[0]*sin(u[1])*cos(u[2]) );
        U.push_back( u[0]*sin(u[2]) );

        // Initialize to u[0]*x[0] polynomial
        dx.clear();

        switch (m_type)
        {
            case inTime:
            {
                // Sanity check
                if (x.size()!=6)
                    ipanema_throw("Input size should be 6");

                // Independent variable is time
                double t = s;

                // Aux
                StateType B2       ( 1.0-pow(x[1],2.0)-pow(x[2],2.0) );
                StateType B        ( sqrt(B2) );
                StateType Bsqrtamu ( sqrt( (1.0-pow(x[1],2.0)-pow(x[2],2.0))*x[0]/m_pParams->mu ) );
                StateType PHI      ( 1.0+x[1]*sin(x[5])+x[2]*cos(x[5]) );
                StateType PHI2     ( pow(1.0+x[1]*sin(x[5])+x[2]*cos(x[5]),2.0) );

                // da/dt
                dx.push_back( 2.0*sqrt( pow(x[0],3.0)/m_pParams->mu/B2 ) );
                dx[0] *=  U[0]*(x[2]*sin(x[5])-x[1]*cos(x[5])) + U[1]*PHI;

                // dP1/dt
                dx.push_back( Bsqrtamu );
                dx[1] *= -U[0]*cos(x[5]) + U[1]*((x[1]+sin(x[5]))/PHI+sin(x[5])) - U[2]*x[2]*(x[3]*cos(x[5])-x[4]*sin(x[5]))/PHI ;

                // dP2/dt
                dx.push_back( Bsqrtamu );
                dx[2] *= +U[0]*sin(x[5]) + U[1]*((x[2]+cos(x[5]))/PHI+cos(x[5])) - U[2]*x[1]*(x[3]*cos(x[5])-x[4]*sin(x[5]))/PHI ;

                // dQ1/dt and dQ2/dt
                //
                // Initialise
                dx.push_back( U[2]*Bsqrtamu/2.0*(1.0+pow(x[3],2.0)+pow(x[4],2.0))/PHI );
                dx.push_back( dx.back() );

                // dQ1/dt
                dx[3] *= sin(x[5]);

                // dQ2/dt
                dx[4] *= cos(x[5]);

                // dL/dt
                dx.push_back( PHI2*sqrt(m_pParams->mu) / pow(x[0]*B2,3.0/2.0)  );

                // Sanity check
                if (dx.size()!=6)
                    ipanema_throw("Output size should be 6");
            }
                break;
            case inLongitude:
            {
                // Sanity check
                if (x.size()!=5)
                    ipanema_throw("Input size should be 5");

                // Independent variable is longitude
                double L = s;

                StateType B2   ( 1.0-pow(x[1],2.0)-pow(x[2],2.0) );
                StateType B4a2 ( pow(1.0-pow(x[1],2.0)-pow(x[2],2.0),2.0)*pow(x[0],2.0) );
                StateType PHI  ( 1.0+x[1]*std::sin(L)+x[2]*std::cos(L) );
                StateType PHI2 ( pow(PHI,2.0) );
                StateType PHI3 ( pow(PHI,3.0) );
                StateType B4a2phi2 ( B4a2/PHI2 );
                StateType B4a2phi3 ( B4a2/PHI3 );

                // da/dL
                dx.push_back( 2.0*B2*pow(x[0],3.0)/m_pParams->mu );
                dx[0] *=  U[0]*(x[2]*std::sin(L)-x[1]*std::cos(L))/PHI2 + U[1]/PHI;

                // dP1/dL and dP2/dL
                //
                // Initialise
                dx.push_back( B4a2phi2/m_pParams->mu );
                dx.push_back( dx.back() );

                // dP1/dL
                dx[1] *= -U[0]*std::cos(L) + U[1]*((x[1]+std::sin(L))/PHI+std::sin(L)) - U[2]*x[2]*(x[3]*std::cos(L)-x[4]*std::sin(L))/PHI ;

                // dP2/dL
                dx[2] *= +U[0]*std::sin(L) + U[1]*((x[2]+std::cos(L))/PHI+std::cos(L)) + U[2]*x[1]*(x[3]*std::cos(L)-x[4]*std::sin(L))/PHI ;

                // dQ1/dL and dQ2/dL
                //
                // Initialise
                dx.push_back( B4a2phi3/(2.0*m_pParams->mu)*(1.0+pow(x[3],2.0)+pow(x[4],2.0))*U[2] );
                dx.push_back( dx.back() );

                // dQ1/dL
                dx[3] *= std::sin(L);

                // dQ2/dL
                dx[4] *= std::cos(L);

                // Sanity check
                if (dx.size()!=5)
                    ipanema_throw("Output size should be 5");

            }
                break;
            default:
                ipanema_throw("type not implemented");
                break;
        }

        return 0;

    } // evaluate

}; // class fableDynamics




/**
 * Sample v_infinity distribution
 * @param[in]  nSamples to generate
 * @param[out] samples generated
 */
int getInitialVInfSamples( vector<vector<double>> &samples, vector<double> &densitySample, const unsigned int nnSamples = nSamples );



/**
 * Convert samples in v_inf to samples in equinoctial elements
 *
 * @param referenceEqX reference state to add v_inf
 * @param vInfSamples
 * @param convertedSamples
 * @return
 */
int convertVInfSamplesToEquinoctial( const vector<double>         &earthEq,
                                     const vector<vector<double>> &vInfSamples,
                                     vector<vector<double>>       &convertedSamples );


/**
 * Get initial guess from file
 */
int getInitialGuess( const std::string      &inputFile,
                     const bool             &getRefinedControl,
                     vector<double>         &earthEqInitial,
                     vector<double>         &sGuess,
                     vector<vector<double>> &xGuess,
                     vector<vector<double>> &uGuess );


/**
 * Reintegrate initial guess
 */
int reintegrateInitialGuess();


/**
 * Forward declataraion
 * @return
 */
int equinoctialOptimisation ();

/**
 * Forward declaration
 * @return
 */
int reintegrateEquinoctial ();


/**
 * Function to get final ellipsoid semi axes lengths
 * @return
 */
int getFinalEllipsoidSemiAxes( const vector<double> &initialLB,
                               const vector<double> &initialUB,
                               vector<double> &finalSigma);


/**
 * Generate sobol samples in domain. Propagate with hard-coded control, and print output file with DR and DV
 * and density of corresponding density function used
 *
 */
int propagateAndExportSobolSamplesWithDifferentDensities();



/**
 * 
 * Main function
 * 
 */
int main(){

    cout << endl << "Welcome to the IPANEMA Planar Low Thrust example!" << endl;

    // Aux
    int check;

    // Reintegrate guess
//    check = reintegrateInitialGuess();

    // Run optimisation
    //check = equinoctialOptimisation();

//    nSamples = 1000;

    // Run a posteriori analysis
    check = reintegrateEquinoctial();

    // Propagate and print samples
    //check = propagateAndExportSobolSamplesWithDifferentDensities();

    cout << endl << "Goodbye from the IPANEMA Planar Low Thrust example!" << endl;

    return 0;
}



/**
 * Collection of density functions
 */
int computeVSampleDensity(const unsigned int           &index,
                          const vector<vector<double>> &vSamples,
                          vector<double>               &density)
{
    // Number of samples
    auto nS = vSamples.size();
    density = vector<double>(nS,1.0);

    switch (index)
    {
        case 0:
        {
            double sigmaVInf = abs(v_infMinDev)/4.0,
                    sigmaAlpha = abs(alpha_infMinDev)/4.0,
                    sigmaBeta = abs(beta_infMinDev)/4.0;

            for (auto i = 0 ; i < nS; i++)
            {
                //density[i] *= gsl_ran_gaussian_tail_pdf(-vSamples[i][0]/sigmaVInf,0.0,1.0);
                //density[i] *= gsl_ran_gaussian_pdf(vSamples[i][1]/sigmaAlpha,1.0);
                //density[i] *= gsl_ran_gaussian_pdf(vSamples[i][2]/sigmaBeta,1.0);
                density[i] *= gsl_ran_gaussian_tail_pdf(-vSamples[i][0],0.0,sigmaVInf);
                density[i] *= gsl_ran_gaussian_pdf(vSamples[i][1],sigmaAlpha);
                density[i] *= gsl_ran_gaussian_pdf(vSamples[i][2],sigmaBeta);
            }
        }
        break;
//        case 1:
//        {
//
//        }
//        break;
        default:
            ipanema_throw("error");
            break;
    }

    // Successful
    return 0;
}



/**
 * Generate sobol samples in domain. Propagate with hard-coded control, and print output file with DR and DV
 * and density of corresponding density function used
 *
 */
int propagateAndExportSobolSamplesWithDifferentDensities()
{
    // Aux
    int check;

    // Get initial guess
    string inputFile ("fableApophis/equinoctial3DFirstGuess.txt");
    vector<double> sGuess, earthEqInitial;
    vector<vector<double>> xGuess, uGuess, uRefined;
    static bool   useRefinedControl = true;
    check = getInitialGuess(inputFile,useRefinedControl,earthEqInitial,sGuess,xGuess,uRefined);

    // Velocity samples
    vector<vector<double>> vSamples(nSamples);

    // GSL sobol generator for velocity errors
    auto nv = 3;
    vector<double> LB(nv), UB(nv);
    LB[0] = v_infMinDev;      UB[0] = 0.0;
    LB[1] = -alpha_infMinDev; UB[1] = alpha_infMinDev;
    LB[2] = -beta_infMinDev;  UB[2] = beta_infMinDev;

    // generate samples
    gsl_qrng* q = gsl_qrng_alloc(gsl_qrng_sobol, nv);
    double* pV = new double[nv];
    for ( auto j = 0 ; j < nSamples; j++ )
    {
        gsl_qrng_get(q,pV);

        vSamples[j] = vector<double>(nv);
        for (auto k = 0; k < nv; ++k)
        {
            vSamples[j][k] = LB[k]+pV[k]*(UB[k]-LB[k]);
        }
    }
    delete [] pV;
    pV = nullptr;
    gsl_qrng_free(q);
	

    // Compute density for several densities
    unsigned int nPdf = 1;
    vector<vector<double>> vSamplesDensities(nPdf);
	
	//vSamples.clear();
	//check = getInitialVInfSamples( vSamples, vSamplesDensities[0] , nSamples );
	
	//for (auto i = 0 ; i < nSamples; i++)
	//{
	//	vSamples[i][0] -= v_inf/L_scale*T_scale;
	//	vSamples[i][0] *= L_scale/T_scale;
	//	vSamples[i][1] -= alpha_inf;
	//	vSamples[i][2] -= beta_inf;
	//}
	//computeVSampleDensity(0,vSamples,vSamplesDensities[1]);
	
	//for (auto i = 0 ; i < nSamples; i++)
	//{
	//	printf("%+5.5f %+5.5f %+5.5f %+5.5f %+5.5f\n",  vSamples[i][0],  vSamples[i][1],  vSamples[i][2], vSamplesDensities[0][i],  vSamplesDensities[1][i]);
	//}
	
	//ipanema_throw("stop");
	
    for (auto i = 0 ; i < nPdf; i++)
        computeVSampleDensity(i,vSamples,vSamplesDensities[i]);

	// Print to file
	vector<vector<double>> initialSamplesToFile(vSamples);
	for (auto i=0; i < nSamples; i++)
		for (auto j = 0 ; j < nPdf; j++)
			initialSamplesToFile[i].push_back(vSamplesDensities[j][i]);
	dlmwrite("../../results/fableApophis/initialVSamplesAndPdf.csv",initialSamplesToFile);
		
		
	//for (auto i = 0 ; i < nSamples; i++)
	//{
	//	printf("%+5.5f %+5.5f %+5.5f %+5.5f \n",  vSamples[i][0],  vSamples[i][1],  vSamples[i][2], vSamplesDensities[0][i]);
	//}
	//ipanema_throw("stop");

    // Shift and scale
    for ( auto j = 0 ; j < nSamples; j++ )
    {
        vSamples[j][0] = (v_inf+vSamples[j][0]) * T_scale / L_scale;
        vSamples[j][1] = (alpha_inf+vSamples[j][1]);
        vSamples[j][2] = (beta_inf+vSamples[j][2]);
    }
	

	
	//ipanema_throw("stop");

   // Convert v samples to equinoctial elements
    vector<vector<double>> eqInitialSamples, finalSamples;
    check = convertVInfSamplesToEquinoctial( earthEqInitial,
                                             vSamples,
                                             eqInitialSamples );


    /**
     * Double Numerical Propagation
     */


    /**
     * Double dynamics and integrator
     */

    // Dynamics parameters
    fableDynamics<double>::fableDynamicsParams dynamicsParams;
    dynamicsParams.mu                   = mu;
    dynamicsParams.pControlCoefficients = &uRefined;
    dynamicsParams.pSGrid               = &sGuess;

    // Dynamics
    fableDynamics<double> *doubleDynamics =
            new fableDynamics<double>(&dynamicsParams,dynamicsType_);

    // Integrator
    smartmath::integrator::rk4<double> doubleIntegrator(doubleDynamics);

    /**
     * Propagate for FABLE control
     */

    auto ns = 6;
    auto ni = sGuess.size()-1;
    vector<vector<double>> eqXF(nSamples), carXF(nSamples,vector<double>(6));
    vector<double> x0(ns), xf(ns);


    /**
     * Propagate for refined control
     */

    for (auto j = 0; j < nSamples; j++) {
		cout << j+1 << "/" << nSamples << endl;

        // Initialise state to
        x0 = eqInitialSamples[j];

        // Integrate for each independent variable interval
        for (auto i = 0; i < ni; i++) {

            double t0 = sGuess[i];
            double tf = sGuess[i+1];

            // Number of integration steps according to size of interval
            auto nIntStep = (int)ceil((double) nint * (tf-t0) /
                                      (sGuess.back() - sGuess[0])) + 1;

            // Integrate
            doubleIntegrator.integrate(t0,tf, nIntStep, x0, xf);
            x0 = xf;
        }

        // Save final equinoctial state
        eqXF[j] = xf;
        if (dynamicsType_==inLongitude)
            eqXF[j].push_back(sGuess.back());

    }

    // Convert to unscaled cartesian
    for (auto j = 0; j < nSamples; j++)
    {
        // Convert to cartesian
        smartastro::astrocore::conversion_coordinates::eq2car(eqXF[j],mu,carXF[j]);

        // Scale back
        for (auto k = 0; k < 3; k++)
            carXF[j][k] *= L_scale;
        for (auto k = 3; k < 6; k++)
            carXF[j][k] *= L_scale/T_scale;
    }

    // Output
    for ( auto i = 0 ; i < nPdf; i++ )
        for ( auto j = 0 ; j < nSamples; j++ )
            carXF[j].push_back(vSamplesDensities[i][j]);
    dlmwrite("../../results/fableApophis/finalCartesianAndPDF.csv",carXF);

    // Successful
    return 0;
}




/**
 * Sample v_infinity distribution
 * @param[in]  nSamples to generate
 * @param[out] samples generated
 */
int getInitialVInfSamples( vector<vector<double>> &samples, vector<double> &densitySample, const unsigned int nnSamples )
{
    // Output
    samples = vector<vector<double>>(nnSamples,vector<double>(3));
    densitySample = vector<double>(nnSamples,1.0);

    // first is reference
    samples[0][0] = v_inf/L_scale*T_scale;
    samples[0][1] = alpha_inf;
    samples[0][2] = beta_inf;
    densitySample[0] = gsl_ran_gaussian_tail_pdf(0.0,0.0,1.0) * gsl_ran_gaussian_pdf(0.0,1.0) * gsl_ran_gaussian_pdf(0.0,1.0);

    // Generate initial samples
    gsl_rng * r = gsl_rng_alloc(gsl_rng_taus);
    gsl_rng_set(r,11);

    // Sample v_inf
    unsigned int cSamples = 1;
    while (cSamples<nnSamples)
    {
        double sigmaVInf = abs(v_infMinDev)/4.0;
        double sample = -1.0*gsl_ran_gaussian_tail(r,0.0,sigmaVInf); // [m/s]

        if (sample<=0.0 && sample >= v_infMinDev)
        {
            // Assign scaled density
            densitySample[cSamples] *= gsl_ran_gaussian_tail_pdf(-sample,0.0,sigmaVInf);

            // Save sample
            samples[cSamples++][0] = (v_inf+sample)/L_scale*T_scale;
        }
    }

    // Sample alpha
    cSamples = 1;
    while (cSamples<nnSamples)
    {
        double sigmaAlpha = abs(alpha_infMinDev)/4.0;
        double sample = gsl_ran_gaussian(r,sigmaAlpha); // [rad]

        if (abs(sample)<=alpha_infMinDev)
        {
            // Assign density
            densitySample[cSamples] *= gsl_ran_gaussian_pdf(sample,sigmaAlpha);

            // Save sample
            samples[cSamples++][1] = (alpha_inf+sample);
        }
    }

    // Sample beta
    cSamples = 1;
    while (cSamples<nnSamples)
    {
        double sigmaBeta = abs(beta_infMinDev)/4.0;
        double sample = gsl_ran_gaussian(r,sigmaBeta); // [rad]

        if (abs(sample)<=beta_infMinDev)
        {
            // Assign density
            densitySample[cSamples] *= gsl_ran_gaussian_pdf(sample,sigmaBeta);

            // Save sample
            samples[cSamples++][2] = (beta_inf+sample);
        }
    }

    // Free r
    gsl_rng_free (r);

    // Successful
    return 0;
}



/**
 * Convert samples in v_inf to samples in equinoctial elements
 *
 * @param referenceEqX reference state to add v_inf
 * @param vInfSamples
 * @param convertedSamples
 * @return
 */
int convertVInfSamplesToEquinoctial( const vector<double>         &earthEq,
                                     const vector<vector<double>> &vInfSamples,
                                     vector<vector<double>>       &convertedSamples )
{
    // Output
    convertedSamples = vector<vector<double>>(nSamples,vector<double>(6));

    // Reference Earth state in cartesian conditions
    vector<double> earthDepartureCar (6);
    smartastro::astrocore::conversion_coordinates::eq2car(earthEq,mu,earthDepartureCar);

    // Spacecraft samples
    vector<double> scDepartureCar (6), scDepartureEq (6);
    for (auto i = 0 ; i < nSamples; i++)
    {
        // Cartesian initial position
        scDepartureCar = earthDepartureCar;
        scDepartureCar[3] += vInfSamples[i][0]*cos(vInfSamples[i][1])*cos(vInfSamples[i][2]);
        scDepartureCar[4] += vInfSamples[i][0]*sin(vInfSamples[i][1])*cos(vInfSamples[i][2]);
        scDepartureCar[5] += vInfSamples[i][0]*sin(vInfSamples[i][2]);

        // Convert to equinoctial
        smartastro::astrocore::conversion_coordinates::car2eq(scDepartureCar,mu,scDepartureEq);

        // Set the samples
        convertedSamples[i] = scDepartureEq;
    }

    // If in longitude, remove last element
    if (dynamicsType_==inLongitude)
        for (auto i = 0 ; i < nSamples; i++)
            convertedSamples[i].pop_back();

//    // Print initial samples
//    cout << "Initial samples: " << endl;
//    for (auto i = 0 ; i < nSamples; i++)
//    {
//        cout << setprecision(15) << convertedSamples[i][0]*L_scale << " ";
//        for (auto j=1;j<6;j++)
//            cout << convertedSamples[i][j] << " ";
//        cout << "   " << (vInfSamples[i][0])*L_scale/T_scale-v_inf << " " << vInfSamples[i][1]-alpha_inf << " " << vInfSamples[i][2]-beta_inf ;
//        cout << endl;
//    }
//    ipanema_throw("end");

    // Successful
    return 0;
}


/**
 * Get initial guess from file
 */
int getInitialGuess( const std::string      &inputFile,
                     const bool             &getRefinedControl,
                     vector<double>         &earthEqInitial,
                     vector<double>         &sGuess,
                     vector<vector<double>> &xGuess,
                     vector<vector<double>> &uGuess )
{
    // Read input file
    vector<vector<double>> inputFABLE;
    dlmread(inputFile,inputFABLE);

    // Earth conditions
    earthEqInitial = vector<double>(6);
    for ( auto i = 0 ; i < 6 ; i++ )
        earthEqInitial[i] = inputFABLE[0][i+1];
    earthEqInitial[0] /= L_scale;

    // Erase initial conditions
    inputFABLE.erase(inputFABLE.begin(),inputFABLE.begin()+1);

    /**
     * Time discretization
     */

    // Number of control discretization intervals
    auto inputD = inputFABLE.size();
    auto ni = inputD-1;

    // Time vector
    vector<double> t_guess (ni+1) ;
    for ( auto i = 0 ; i < inputD; i++ )
        t_guess[i] = inputFABLE[i][0] / T_scale ; // Scale time originally in s

    /**
     * L discretization
     */

    // L vector
    vector<double> L_guess (ni+1) ;
    for ( auto i = 0 ; i < inputD; i++ )
        L_guess[i] = inputFABLE[i][6];

    /**
     * S discretisation
     */
    if (dynamicsType_==inLongitude)
        sGuess = L_guess;
    else if (dynamicsType_==inTime)
        sGuess = t_guess;

    /**
     * State guess
     */
    // State variables
    unsigned int ns;
    if (dynamicsType_==inLongitude)
        ns = 5;
    else if (dynamicsType_==inTime)
        ns = 6;
    else {
        ipanema_throw("error");
        ns = 0;
    }

    // State guess
    xGuess = vector<vector<double>> (ni+1,vector<double>(ns));
    for ( auto i = 0 ; i < inputD; i++ )
    {
        for (auto j = 0; j < ns; j++)
        {
            xGuess[i][j] = inputFABLE[i][j+1];
        }

        // Scale semi-major axis originally in m
        xGuess[i][0] /= L_scale;
    }


    /**
     * Control parameter guess
     */
    uGuess = vector<vector<double>>(ni,vector<double>(nc,0.0));

    // Use already optimised control
    string inputRefinedControl = "../../results/fableApophis/refinedControlDeg" +
                                 std::to_string(nDeg) +".txt";

    if (getRefinedControl)
    {
        dlmread(inputRefinedControl,uGuess);
    }
    else
    {
        for ( auto i = 0 ; i < ni; i++ )
            for (auto j = 0; j < nc; j++)
                uGuess[i][j] = inputFABLE[i][j+7];
    }

    for ( auto i = 0 ; i < ni; i++ )
        uGuess[i][0] *= pow(T_scale,2.0)/L_scale; // originally [m/s^2]

    // Successful
    return 0;
}



int reintegrateInitialGuess()
{

    // Aux
    int check;

    // State variables
    unsigned int ns;
    if (dynamicsType_==inLongitude)
        ns = 5;
    else if (dynamicsType_==inTime)
        ns = 6;
    else {
        ipanema_throw("error");
        ns = 0;
    }

    /**
     * Read input guesses
     */

    // Read input file by FABLE
    string inputFile ("fableApophis/equinoctial3DFirstGuess.txt");
    vector<double> sGuess, earthEqInitial;
    vector<vector<double>> xGuess, uGuess;
    // Use already computed control
    static bool   useRefinedControl = false;
    check = getInitialGuess(inputFile,useRefinedControl,earthEqInitial,sGuess,xGuess,uGuess);

    // Number of discretisation intervals (input)
    auto ni = sGuess.size()-1;

    /**
     * Initial and final samples [according to initial v_inf]
     */

    /**
     * Initialise dynamics and integrator
     */

    // Dynamics parameters
    fableDynamics<double>::fableDynamicsParams dynamicsParams;
    dynamicsParams.mu                   = mu;
    dynamicsParams.pControlCoefficients = &uGuess;
    dynamicsParams.pSGrid               = &sGuess;

    // Dynamics
    fableDynamics<double>* doubleDynamics =
            new fableDynamics<double>(&dynamicsParams,dynamicsType_);

    // Integrator
    smartmath::integrator::rk4<double> doubleIntegrator (doubleDynamics);

    // Propagate interval by interval
    vector<vector<double>> histX(ni+1);
    histX[0] = xGuess[0];

    // Number of integration steps for unit s
    double nStepPerUnitS = ((double)nint)/(sGuess.back()-sGuess[0]) ;

    // Number of integration steps per unit
//    ni = 1;
    for (auto i = 0 ; i < ni; i++)
    {
        // Number of integration steps for current interval
        auto nStep_i = (int)ceil( nStepPerUnitS*(sGuess[i+1]-sGuess[i]) );
//        cout << "Iter " << i+1 << "/" << ni << "; nStep = " << nStep_i << endl;
        doubleIntegrator.integrate(sGuess[i],sGuess[i+1],nStep_i,histX[i],histX[i+1]);
    }


    /**
     * Print reference and reintegrated
     */
    cout << "Control guess:" << endl;
    for (auto i = 0; i < ni; i++) {

        printf("%+5.10f ",sGuess[i]);

        for (auto j = 0; j < nc; j++)
            printf("%+5.10f ",uGuess[i][j]);
        cout << endl;
    }
    cout << endl;


    cout << "Reference:" << endl;
    for (auto i = 0; i <= ni; i++) {

        printf("%+5.10f ",sGuess[i]);

        for (auto j = 0; j < ns; j++)
            printf("%+5.10f ",xGuess[i][j]);

//        if (dynamicsType_==inLongitude)
//            printf("%+5.10f ",L_guess[i] );
        cout << endl;
    }
    cout << endl;

    cout << "ReIntegrated:" << endl;
    for (auto i = 0; i <= ni; i++) {

        printf("%+5.10f ",sGuess[i]);

        for (auto j = 0; j < ns; j++)
            printf("%+5.10f ",histX[i][j]);

//        if (dynamicsType_==inLongitude)
//            printf("%+5.10f ",L_guess[i] );
        cout << endl;
    }
    cout << endl;

    cout << "Difference:" << endl;
    for (auto i = 0; i <= ni; i++) {

        printf("%+5.10f ",sGuess[i]);

        for (auto j = 0; j < ns; j++)
            printf("%+5.10f ",histX[i][j]-xGuess[i][j]);

//        if (dynamicsType_==inLongitude)
//            printf("%+5.10f ",L_guess[i] );
        cout << endl;
    }
    cout << endl;


    // Successful
    return 0;
}



/**
 *
 * @return
 */
int equinoctialOptimisation () {

    // Aux
    int check;


    // State variables
    unsigned int ns;
    if (dynamicsType_==inLongitude)
        ns = 5;
    else if (dynamicsType_==inTime)
        ns = 6;
    else {
        ipanema_throw("error");
        ns = 0;
    }

    // Polynomial settings
    int nVar = ns+nc;

    // Scales
    vector<double> controlScale(nc,1.0);
    vector<double> constraintScale (ng,1.0);
    double objectiveScale = 10.0;

    /**
     * Read input guesses
     */

    // Read input file by FABLE
    string inputFile ("fableApophis/equinoctial3DFirstGuess.txt");
    vector<double> sGuess, earthEqInitial;
    vector<vector<double>> xGuess, uGuess;
    bool useRefinedControl = true;
    check = getInitialGuess(inputFile,useRefinedControl,earthEqInitial,sGuess,xGuess,uGuess);

    // Number of discretisation intervals (input)
    auto ni = sGuess.size()-1;





    /**
     * Initial and final samples [according to initial v_inf]
     */
    vector<vector<double>> vInfSamples, eqInitialSamples, finalSamples;
    vector<double> densitySample;
    check = getInitialVInfSamples( vInfSamples, densitySample , nSamples );
    check = convertVInfSamplesToEquinoctial( earthEqInitial,
                                             vInfSamples,
                                             eqInitialSamples );

    // Sample + density
    vector<vector<double>> outputInitialSamples(eqInitialSamples);
    for (auto i = 0 ; i < nSamples; i++)
        outputInitialSamples[i].push_back(densitySample[i]);

    dlmwrite("../../results/fableApophis/initialSamples.txt",outputInitialSamples);

//    // Print initial samples
//    cout << "Initial samples: " << endl;
//    for (auto i = 0 ; i < nSamples; i++)
//    {
//        cout << setprecision(15) << eqInitialSamples[i][0] << " ";
//        for (auto j=1;j<ns;j++)
//            cout << eqInitialSamples[i][j] << " ";
//        if (dynamicsType_==inLongitude)
//            cout << L_guess[0] << " ";
//        cout << "   " << (vInfSamples[i][0])*L_scale/T_scale-v_inf << " " << vInfSamples[i][1]-alpha_inf << " " << vInfSamples[i][2]-beta_inf ;
//        cout << endl;
//    }





    /**
     *
     *
     * Optimiser reference quantities
     *
     *
     */





    /**
     * Control parameter bounds [simplified]
     */

    vector<vector<double>> realUBounds(nc,vector<double>(2));
    double deltaControl = 1.0e-4;

    // Magnitude
    realUBounds[0][0] = 0.0-deltaControl;
    realUBounds[0][1] = maxT/m0+deltaControl;

    // Alpha
    realUBounds[1][0] = -3.0;
    realUBounds[1][1] = -0.17;

    // Beta
    realUBounds[2][0] = -0.5;
    realUBounds[2][1] = +0.5;

    // Dynamics real bounds (cause worhp checks slightly more)
    vector<vector<double>> largerUBounds(realUBounds);
    for (auto i = 0 ; i < nc; i++)
    {
        largerUBounds[i][0] -= 1.0e-4;
        largerUBounds[i][1] += 1.0e-4;
    }



    /**
     * Initial and final boundary conditions
     */

    vector<double> refInitialCondition (ns);
    vector<double> refFinalCondition   (ns);

    // Guess assumed as perfect
    refInitialCondition = xGuess[0];
    refFinalCondition   = xGuess[ni];




    /**
     * Initial starting box zone
     */

    // Define relative to reference value magnitude
    vector<double> initialMean(refInitialCondition),
            initialLB(ns,+1.0e3),
            initialUB(ns,-1.0e3);

    // Custom
    initialLB = eqInitialSamples[0];
    initialUB = eqInitialSamples[0];
    for (auto i = 1 ; i < nSamples; i++)
    {
        for (auto j = 0 ; j < ns; j++)
        {
            if (initialLB[j]>eqInitialSamples[i][j])
                initialLB[j] = eqInitialSamples[i][j];
            if (initialUB[j]<eqInitialSamples[i][j])
                initialUB[j] = eqInitialSamples[i][j];
        }
    }

    // Margin
    initialLB[0] -= abs(initialLB[0])*0.001;
    initialUB[0] += abs(initialUB[0])*0.001;
    for (auto j = 1 ; j < ns; j++)
    {
        initialLB[j] -= abs(initialLB[j])*0.01;
        initialUB[j] += abs(initialUB[j])*0.01;
    }

    // Sanity check
    for (auto j = 0 ; j < ns; j++)
        if (initialLB[j]>=initialUB[j])
            ipanema_throw("Lower bound larger than upper bound");

    /**
     * Define final target ellipsoid dimension
     */

    // Final mean and sigma
    vector<double> finalMean (refFinalCondition),
            finalSigma(ns),
            finalLB(ns),
            finalUB(ns);

    // Get final sigma
    getFinalEllipsoidSemiAxes(initialLB,initialUB,finalSigma);

    // Symmetric final Box
    finalLB = finalMean-finalSigma;
    finalUB = finalMean+finalSigma;



    // Print output mean, LB, UB for plot
    vector<vector<double>> initialFinalMeanLBUBOutput (2,vector<double>(3*ns));
    // Initial mean - LB - UB
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i]      = initialMean[i];
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i+ns]   = initialLB[i];
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i+2*ns] = initialUB[i];
    // Final mean - LB - UB
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i]      = finalMean[i];
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i+ns]   = finalLB[i];
    for (auto i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i+2*ns] = finalUB[i];
    dlmwrite("../../results/fableApophis/initialFinalMeanLBUB.txt",initialFinalMeanLBUBOutput);
    initialFinalMeanLBUBOutput.clear();





    /**
     * Initial polynomial box
     */

    vector<smartuq::polynomial::chebyshev_polynomial> initialPolynomials;
    for (auto i = 0 ; i < ns; i++ )
    {
        initialPolynomials.emplace_back(
                smartuq::polynomial::chebyshev_polynomial(nVar,nDeg,i,
                                                          initialLB[i],
                                                          initialUB[i],
                                                          monomial)
        );
    }



    /**
     * Polynomial control
     */

    // The same for all the times
    vector<smartuq::polynomial::chebyshev_polynomial> polynomialControl_i;
    for ( auto i = 0 ; i < nc; i++ )
    {
        polynomialControl_i.emplace_back(
                smartuq::polynomial::chebyshev_polynomial(nVar,nDeg,i+ns,
                                                          largerUBounds[i][0],
                                                          largerUBounds[i][1],
                                                          monomial)
        );
    }

    // Same for all discretisation interval
    vector<vector<smartuq::polynomial::chebyshev_polynomial>> polynomialControl (ni,polynomialControl_i);





    /**
     * Quantities to be optimised
     * (initialised to guess)
     */
    vector<double> sRefined (sGuess) ;
    vector<vector<double>> uRefined(uGuess);


    /**
     * Initialise dynamics and integrator
     */

    // Dynamics parameters
    fableDynamics<smartuq::polynomial::chebyshev_polynomial>::fableDynamicsParams dynamicsParams;
    dynamicsParams.mu                   = mu;
    dynamicsParams.pControlCoefficients = &polynomialControl;
    dynamicsParams.pSGrid               = &sRefined;

    // Dynamics
    fableDynamics<smartuq::polynomial::chebyshev_polynomial>* polynomialDynamics =
            new fableDynamics<smartuq::polynomial::chebyshev_polynomial>(&dynamicsParams,dynamicsType_);

    // Integrator
    smartmath::integrator::rk4<smartuq::polynomial::chebyshev_polynomial> polynomialIntegrator (polynomialDynamics);






    /**
     * Auxiliary quantities to be shared through pointer in several classes
     */
    bool majorIteratior = false; // Belongs to constraint and NLP solver


    /**
     * Surrogate propagator
     */

    // Parameters
    intrusivePolynomialPropagator::intrusivePolynomialPropagatorParams propagatorParams;

    propagatorParams.pStartPolynomials    = &initialPolynomials;
    propagatorParams.pIntegrator          = &polynomialIntegrator;
    propagatorParams.pTimeGrid            = &sRefined;
    propagatorParams.pControlCoefficients = &uRefined;
    propagatorParams.controlBounds        = largerUBounds;
    propagatorParams.nIntegrationSteps    = nint;
    propagatorParams.isSystemAutomonous   = false;
    propagatorParams.printToFile          = true;
    propagatorParams.outputFilePrefix     = "../../results/fableApophis/";
    propagatorParams.readFromFile         = true;
    propagatorParams.inputFilePrefix      = propagatorParams.outputFilePrefix;

    // Predefined range for polynomial propagation
    propagatorParams.usePredefinedRange   = true;
    propagatorParams.predefinedRange      = vector<vector<double>>(ns,vector<double>(2,0.0));

    // a [au]
    propagatorParams.predefinedRange[0][0] = +0.89;
    propagatorParams.predefinedRange[0][1] = +1.05;
    // p1 [rad]
    propagatorParams.predefinedRange[1][0] = -0.15;
    propagatorParams.predefinedRange[1][1] = -0.05;
    // p2 [rad]
    propagatorParams.predefinedRange[2][0] = +0.05;
    propagatorParams.predefinedRange[2][1] = +0.20;
    // q1 [rad]
    propagatorParams.predefinedRange[3][0] = -0.017;
    propagatorParams.predefinedRange[3][1] = -0.007;
    // q2 [rad]
    propagatorParams.predefinedRange[4][0] = -0.030;
    propagatorParams.predefinedRange[4][1] = -0.020;

    if (dynamicsType_==inTime)
    {
        ipanema_throw("don't know how to do");

        // L [rad]
        propagatorParams.predefinedRange[5][0] = +0.05;
        propagatorParams.predefinedRange[5][1] = +0.20;
    }

    // Initialise surrogate propagator
    intrusivePolynomialPropagator polynomialPropagator(&propagatorParams);

    // Explicitly build polynomial surrogate
    polynomialPropagator.build();

//    ipanema_throw("stop");







    /**
     *
     * Construct solver
     *
     */



    /**
     * Final probabilistic constraint
     */


    // Worhp bounds
    vector<double> constraintWorhpBounds(2);
    constraintWorhpBounds[0] = -1.0e21;
    constraintWorhpBounds[1] = 5.0e-2;//4.7e-2;


    // Parameters
    samplingProbabilityConstraint::samplingProbabilityConstraintParams constraintParams;

    constraintParams.pSamples           = &finalSamples;
    constraintParams.sampleSubsetOption = false;
    constraintParams.expectationMC      = true; // Sampled according to initial distribution
    constraintParams.pUseFullSampleSet  = &majorIteratior;
    constraintParams.membershipType     = samplingProbabilityConstraint::BIQUADRATIC_MEMBERSHIP;
    constraintParams.setValueLB         = 0.0;
    constraintParams.setValueUB         = 1.0;

    // Final condition
    if (finalConditionInEq)
    {
        // Ellipsoid in equinoctial
        constraintParams.pSetCondition =
                [mean = finalMean, sigma = finalSigma](const std::vector<double> &x) -> double {
                    unsigned int ns = x.size();
                    double y = 0.0;

                    // Evaluate ellipsoid contour
                    for (auto j = 0; j < ns; j++) {
                        y += std::pow((x[j] - mean[j]) / sigma[j], 2.0);
                    }

                    return y;
                };
    }
    else
    {
        // Cartesian check
        constraintParams.pSetCondition      =
                [mean=finalMean,mmaxDR=maxDR,mmaxDV=maxDV,Lf=sGuess.back()] (const std::vector<double>& x) -> double {

                    // Convert sample in equinoctial
                    vector<double> eq(x),car(6);
                    if (eq.size()==5)
                        eq.push_back(Lf);
                    else
                        ipanema_throw("error");

                    vector<double> eqTarget(mean), carTarget(6);
                    eqTarget.push_back(Lf);

                    // Convert to cartesian
                    smartastro::astrocore::conversion_coordinates::eq2car(eq,mu,car);
                    smartastro::astrocore::conversion_coordinates::eq2car(eqTarget,mu,carTarget);

                    // Compute DR and DV
                    vector<double> dx(car-carTarget);
                    double DR = sqrt( pow(dx[0],2.0)+pow(dx[1],2.0)+pow(dx[2],2.0) );
                    double DV = sqrt( pow(dx[3],2.0)+pow(dx[4],2.0)+pow(dx[5],2.0) );

                    double y = 10.0;
                    if ( (DR<=mmaxDR) && (DV<=mmaxDV) )
                        y = 0.5;

                    return y;
                };
    }


    // Initialise constraint
    samplingProbabilityConstraint finalConstraint(&constraintParams);




    /**
     * Objective class
     */

    // Params
    integralOfControl::integralOfControlParams objectiveParams;

    objectiveParams.pTimeGrid              = &sGuess;
    objectiveParams.pControlCoefficients   = &uRefined;
    objectiveParams.objectiveIntegrandType = integralOfControl::MODULUS_OBJECTIVE;
    objectiveParams.controlType            = integralOfControl::MAGNITUDEANGLES_CONTROL;

    // Initialise objective
    integralOfControl objective(&objectiveParams);



    /**
     * Ipanema solver
     */

    // Parameters
    ipanemaSolver::ipanemaSolverParams solverParams;

    solverParams.pControlCoefficients   = &uRefined;
    solverParams.pTimeGrid              = &sGuess;
    solverParams.pInitialSamples        = &eqInitialSamples;
    solverParams.pFinalSamples          = &finalSamples;
    solverParams.pPropagator            = bind(&intrusivePolynomialPropagator::propagate,
                                               &polynomialPropagator,
                                               std::placeholders::_1);
    solverParams.pObjective             = bind(&integralOfControl::evaluate,
                                               &objective);
    solverParams.pConstraints           = vector<function<double()>>(1,bind(&samplingProbabilityConstraint::evaluate,
                                                                            &finalConstraint));
    solverParams.constraintsBounds      = vector<vector<double>>(ng,constraintWorhpBounds);
    solverParams.controlBounds          = realUBounds;
    solverParams.constraintScale        = constraintScale;
    solverParams.controlScale           = controlScale;
    solverParams.pMajorIteration        = &majorIteratior;
    solverParams.objectiveScale         = objectiveScale;
    solverParams.constraintRelax        = false;
    solverParams.gRelaxPenalty          = vector<double>(ng,1.0);
    solverParams.gRelaxUB               = 0.0;
    solverParams.optimiseFromFeasible   = true;
    solverParams.onlyFeasible           = true;

    // Initialise solver
    ipanemaSolver solver(&solverParams);



    /**
     * Print ms dimension
     */

    cout << "Number of multiple shooting states = " << ns << endl;
    cout << "Number of multiple shooting controls = " << nc << endl;
    cout << "Number of multiple shooting intervals = " << ni << endl;
    cout << "Number of ipanema polynomial variables = " << nVar << endl;
    cout << "Degree of ipanema polynomial = " << nDeg << endl;


    /**
     * Solve by WORHP
     */

    solver.solveByWorhp();


    /**
     * Print final quantities
     */

    // Final free vector
    vector<double> x_refined = solver.get_x();
    vector<vector<double>> xFreeVariables (1,x_refined);

    // Save control
    vector<vector<double>> outputControl (ni,vector<double>(nc));
    vector<vector<double>> guessControl (ni,vector<double>(nc));
    for ( auto i = 0 ; i < ni; i++ )
    {
        for ( auto j = 0 ; j < nc; j++ )
        {
            outputControl[i][j] = uRefined[i][j];
            guessControl[i][j]  = uGuess[i][j];
        }
        outputControl[i][0] /= pow(T_scale,2.0)/L_scale;
        guessControl[i][0] /= pow(T_scale,2.0)/L_scale;
    }

    cout << "Final free vector: " ;
    for ( auto i = 0 ; i < x_refined.size(); i++ )
        cout << x_refined[i] << " " ;
    cout << endl;

    dlmwrite("../../results/fableApophis/refinedControlDeg" +
             std::to_string(nDeg) +".txt",outputControl);
//    dlmwrite("../../results/fableApophis/fableControlDeg" +
//             std::to_string(nDeg) +".txt",guessControl);

    double finalObjControl = solver.get_J();
    if(solverParams.constraintRelax)
        finalObjControl -= solverParams.gRelaxPenalty[0]*x_refined.back();
    cout << "Final control objective unscaled  = " << finalObjControl*objectiveScale << endl;

}











/**
 * Function to get final ellipsoid semi axes lengths
 * @return
 */
int getFinalEllipsoidSemiAxes( const vector<double> &initialLB,
                               const vector<double> &initialUB,
                               vector<double> &finalSigma)
{
    // Define relative to reference value magnitude
    finalSigma = (initialUB-initialLB)/2.0;
    finalSigma[0] /= 2.0; // Custom a-sigma
    finalSigma[1] /= 1.5; // Custom p1-sigma
    finalSigma[3] /= 1.5; // Custom q1-sigma
    finalSigma[4] /= 1.5; // Custom q2-sigma

    // Successful
    return 0;
}




















int reintegrateEquinoctial () {


    /**
     * Dimension of the problem
     */

    // State variables
    unsigned int ns;
    if (dynamicsType_==inLongitude)
        ns = 5;
    else if (dynamicsType_==inTime)
        ns = 6;

    // Polynomial settings
    int nVar = ns + nc;


    // Scales
    vector<double> controlScale(nc, 1.0);
    vector<double> constraintScale(ng, 1.0);


    /**
     * Read input guesses
     */

    // Read input file by FABLE
    string inputFile ("fableApophis/equinoctial3DFirstGuess.txt");
    vector<double> sGuess, earthEqInitial;
    vector<vector<double>> xGuess, uGuess;
    bool useRefinedControl = false;
    int check = getInitialGuess(inputFile,useRefinedControl,earthEqInitial,sGuess,xGuess,uGuess);

    // Refined control
    vector<vector<double>> uRefined;
    dlmread("../../results/fableApophis/refinedControlDeg" + std::to_string(nDeg) + ".txt", uRefined);
    for ( vector<double> &v : uRefined )
        v[0] *= pow(T_scale,2.0)/L_scale; // originally [m/s^2]

    // Initial and final samples [according to initial v_inf]
    vector<vector<double>> vInfSamples, eqInitialSamples, finalSamples;
    vector<double> densitySample;
    check = getInitialVInfSamples( vInfSamples, densitySample, nSamples );
    check = convertVInfSamplesToEquinoctial( earthEqInitial,
                                             vInfSamples,
                                             eqInitialSamples );

    // Sample + density
    vector<vector<double>> outputInitialSamples(eqInitialSamples);
    for (auto i = 0 ; i < nSamples; i++)
        outputInitialSamples[i].push_back(densitySample[i]);

    dlmwrite("../../results/fableApophis/reintegrateInitialSamples.txt",outputInitialSamples);


    // Size
    auto ni = sGuess.size()-1;

    cout << endl << "uGuess = " << endl;
    for (auto i = 0; i < ni; i++)
        printf("%+5.5f %+5.5f %+5.5f\n",uGuess[i][0],uGuess[i][1],uGuess[i][2]);

    cout << endl << "uRefined = " << endl;
    for (auto i = 0; i < ni; i++)
        printf("%+5.5f %+5.5f %+5.5f\n",uRefined[i][0],uRefined[i][1],uRefined[i][2]);



    /**
     * Initial and final boundary conditions
     */

    vector<double> refInitialCondition (ns);
    vector<double> refFinalCondition (ns);

    refInitialCondition = xGuess[0];
    refFinalCondition   = xGuess.back();





    /**
     * Initial starting box zone
     */


    // Define relative to reference value magnitude
    vector<double> initialMean(refInitialCondition),
            initialLB(ns,+1.0e3),
            initialUB(ns,-1.0e3);

    // Samples of optimisation
    vector<vector<double>> optimisedInitialSamples;//(eqInitialSamples);
    dlmread("../../results/fableApophis/initialSamples.txt",optimisedInitialSamples);


    // Remove density
    for ( vector<double>&sample : optimisedInitialSamples )
        sample.pop_back();

    // Custom
    initialLB = optimisedInitialSamples[0];
    initialUB = optimisedInitialSamples[0];
    for (auto i = 1 ; i < optimisedInitialSamples.size(); i++)
    {
        for (auto j = 0 ; j < ns; j++)
        {
            if (initialLB[j]>optimisedInitialSamples[i][j])
                initialLB[j] = optimisedInitialSamples[i][j];
            if (initialUB[j]<optimisedInitialSamples[i][j])
                initialUB[j] = optimisedInitialSamples[i][j];
        }
    }

    // Margin
    initialLB[0] -= abs(initialLB[0])*0.001;
    initialUB[0] += abs(initialUB[0])*0.001;
    for (auto j = 1 ; j < ns; j++)
    {
        initialLB[j] -= abs(initialLB[j])*0.01;
        initialUB[j] += abs(initialUB[j])*0.01;
    }

//    for (auto j = 0 ; j < ns; j++)
//        printf("%+5.5f %+5.5f \n",initialLB[j],initialUB[j]);


    /**
     * Define final target ellipsoid dimension
     */

    // Final mean and sigma
    vector<double> finalMean (refFinalCondition),
            finalSigma(ns),
            finalLB(ns),
            finalUB(ns);

    // Get final sigma
    getFinalEllipsoidSemiAxes(initialLB,initialUB,finalSigma);

    // Symmetric final Box
    finalLB = finalMean-finalSigma;
    finalUB = finalMean+finalSigma;



    /**
     * Control parameter bounds [simplified]
     */

    vector<vector<double>> realUBounds(nc,vector<double>(2));
    double deltaControl = 1.0e-4;

    // Magnitude
    realUBounds[0][0] = 0.0-deltaControl;
    realUBounds[0][1] = maxT/m0+deltaControl;

    // Alpha
    realUBounds[1][0] = -3.0;
    realUBounds[1][1] = -0.17;

    // Beta
    realUBounds[2][0] = -0.5;
    realUBounds[2][1] = +0.5;

    // Dynamics real bounds (cause worhp checks slightly more)
    vector<vector<double>> largerUBounds(realUBounds);
    for (auto i = 0 ; i < nc; i++)
    {
        largerUBounds[i][0] -= 1.0e-4;
        largerUBounds[i][1] += 1.0e-4;
    }





    /**
     * Initial polynomial box
     */

    vector<smartuq::polynomial::chebyshev_polynomial> initialPolynomials;
    for (auto i = 0; i < ns; i++) {
        initialPolynomials.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar, nDeg, i,
                                                          initialLB[i],
                                                          initialUB[i],
                                                          monomial)
        );
    }


    /**
     * Polynomial control
     */

    // The same for all the times
    vector<smartuq::polynomial::chebyshev_polynomial> polynomialControl_i;
    for (int i = 0; i < nc; i++) {
        polynomialControl_i.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar, nDeg, i + ns,
                                                          largerUBounds[i][0],
                                                          largerUBounds[i][1],
                                                          monomial)
        );
    }

    // Same for all discretisation interval
    vector<vector<smartuq::polynomial::chebyshev_polynomial>> polynomialControl(ni, polynomialControl_i);




    /**
     * Final probabilistic constraint
     */

//    vector<double> constraintWorhpBounds(2);
//    constraintWorhpBounds[0] = -1.0e21;
//    constraintWorhpBounds[1] = 0.0;

    // Parameters
    samplingProbabilityConstraint::samplingProbabilityConstraintParams constraintParams;

    constraintParams.sampleSubsetOption = false;
    constraintParams.expectationMC      = true; // Sampled according to initial distribution
    constraintParams.membershipType     = samplingProbabilityConstraint::INDICATOR_MEMBERSHIP;
    constraintParams.setValueLB         = 0.0;
    constraintParams.setValueUB         = 1.0;

    // Final condition
    if (finalConditionInEq)
    {
        // Ellipsoid in equinoctial
        constraintParams.pSetCondition =
                [mean = finalMean, sigma = finalSigma](const std::vector<double> &x) -> double {
                    unsigned int ns = x.size();
                    double y = 0.0;

                    // Evaluate ellipsoid contour
                    for (auto j = 0; j < ns; j++) {
                        y += std::pow((x[j] - mean[j]) / sigma[j], 2.0);
                    }

                    return y;
                };
    }
    else
    {
        // Cartesian check
        constraintParams.pSetCondition      =
                [mean=finalMean,mmaxDR=maxDR,mmaxDV=maxDV,Lf=sGuess.back()] (const std::vector<double>& x) -> double {

                    // Convert sample in equinoctial
                    vector<double> eq(x),car(6);
                    if (eq.size()==5)
                        eq.push_back(Lf);
                    else
                        ipanema_throw("error");

                    vector<double> eqTarget(mean), carTarget(6);
                    eqTarget.push_back(Lf);

                    // Convert to cartesian
                    smartastro::astrocore::conversion_coordinates::eq2car(eq,mu,car);
                    smartastro::astrocore::conversion_coordinates::eq2car(eqTarget,mu,carTarget);

                    // Compute DR and DV
                    vector<double> dx(car-carTarget);
                    double DR = sqrt( pow(dx[0],2.0)+pow(dx[1],2.0)+pow(dx[2],2.0) );
                    double DV = sqrt( pow(dx[3],2.0)+pow(dx[4],2.0)+pow(dx[5],2.0) );

                    double y = 10.0;
                    if ( (DR<=mmaxDR) && (DV<=mmaxDV) )
                        y = 0.5;

                    return y;
                };
    }


    // Initialise constraint
    samplingProbabilityConstraint finalConstraint(&constraintParams);




    /**
     * Initialise polynomial dynamics and integrator
     */

    // Dynamics parameters
    fableDynamics<smartuq::polynomial::chebyshev_polynomial>::fableDynamicsParams polDynamicsParams;
    polDynamicsParams.mu                   = mu;
    polDynamicsParams.pControlCoefficients = &polynomialControl;
    polDynamicsParams.pSGrid               = &sGuess;

    // Dynamics
    fableDynamics<smartuq::polynomial::chebyshev_polynomial> *polynomialDynamics =
            new fableDynamics<smartuq::polynomial::chebyshev_polynomial>(&polDynamicsParams,dynamicsType_);

    // Integrator
    smartmath::integrator::rk4<smartuq::polynomial::chebyshev_polynomial> polynomialIntegrator(polynomialDynamics);





    /**
     * Surrogate propagator
     */

    // Parameters
    intrusivePolynomialPropagator::intrusivePolynomialPropagatorParams propagatorParams;

    propagatorParams.pStartPolynomials    = &initialPolynomials;
    propagatorParams.pIntegrator          = &polynomialIntegrator;
    propagatorParams.pTimeGrid            = &sGuess;
    propagatorParams.pControlCoefficients = &uGuess;
    propagatorParams.controlBounds        = largerUBounds;
    propagatorParams.nIntegrationSteps    = nint;
    propagatorParams.isSystemAutomonous   = false;
    propagatorParams.printToFile          = true;
    propagatorParams.outputFilePrefix     = "../../results/fableApophis/";
    propagatorParams.readFromFile         = true;
    propagatorParams.inputFilePrefix      = propagatorParams.outputFilePrefix;

    // Predefined range for polynomial propagation
    propagatorParams.usePredefinedRange   = false;

    // Initialise surrogate propagator
    intrusivePolynomialPropagator polynomialPropagator(&propagatorParams);

    // Explicitly build polynomial surrogate
    polynomialPropagator.build();


    /**
     * Set of samples to propagate for check
     */

    vector<vector<double>> doubleFinalSamplesFable(nSamples,
                                                   vector<double>(ns));
    vector<vector<double>> polFinalSamplesFable(nSamples,
                                                vector<double>(ns));
    vector<vector<double>> doubleFinalSamplesRefined(nSamples,
                                                     vector<double>(ns));
    vector<vector<double>> polFinalSamplesRefined(nSamples,
                                                  vector<double>(ns));


    /**
     * Double Numerical Propagation
     */

    cout << endl << ">>>>>>>>>> Check double propagation <<<<<<<<<<" << endl;



    /**
     * Double dynamics and integrator
     */

    // Dynamics parameters
    fableDynamics<double>::fableDynamicsParams dynamicsParams;
    dynamicsParams.mu                   = mu;
    dynamicsParams.pControlCoefficients = &uGuess;
    dynamicsParams.pSGrid               = &sGuess;

    // Dynamics
    fableDynamics<double> *doubleDynamics =
            new fableDynamics<double>(&dynamicsParams,dynamicsType_);

    // Integrator
    smartmath::integrator::rk4<double> doubleIntegrator(doubleDynamics);

    /**
     * Propagate for FABLE control
     */

    vector<vector<double>> eqX0F(nSamples),carX0F(nSamples);
    vector<double> x0(ns), xf(ns);

    // For fable solution
    vector<vector<vector<double>>> samplesDoubleFable;
    vector<double> doubleYFable(nSamples);
    for (auto j = 0; j < nSamples; j++) {

        // Initialise state to
        x0 = eqInitialSamples[j];

        // History of each sample
        vector<vector<double>> histSample;
        histSample.push_back(x0);

        // Remove propagated values and Push back in vector for output
        eqX0F[j] = x0;
        if (dynamicsType_==inLongitude)
            eqX0F[j].push_back(sGuess[0]);

        // Integrate for each independent variable interval
        for (auto i = 0; i < ni; i++) {

            // Number of integration steps according to size of interval
            auto nIntStep = (int)ceil(
                    (double)nint*(sGuess[i+1]-sGuess[i]) /
                    (sGuess.back()-sGuess[0]))+1;

            // Integrate
            doubleIntegrator.integrate(sGuess[i],sGuess[i+1],nIntStep,x0,xf);
            x0 = xf;

            // Save for output
            for (auto k = 0; k < ns; k++)
                eqX0F[j].push_back(xf[k]);
            if (dynamicsType_==inLongitude)
                eqX0F[j].push_back(sGuess[i+1]);

            // Save history
            histSample.push_back(xf);
        }

        doubleYFable[j] = constraintParams.pSetCondition(xf);
        doubleFinalSamplesFable[j] = xf;

        // Save history of sample
        samplesDoubleFable.push_back(histSample);
    }
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesFABLEDouble.txt", eqX0F);
    eqX0F.clear();



    /**
     * Propagate for refined control
     */

    // Set control parameters
    dynamicsParams.pControlCoefficients = &uRefined;

    // For refined solution
    vector<vector<vector<double>>> samplesDoubleRefined;
    vector<double> doubleYRefined(nSamples);
    eqX0F = vector<vector<double>>(nSamples);

    // Save output for L-evolution
    auto nTimeEvolutionDiv = 10;
    vector<vector<double>> timeEvolutionRefined(nSamples);

    for (auto j = 0; j < nSamples; j++) {

        // Initialise state to
        x0 = eqInitialSamples[j];

        // History of each sample
        vector<vector<double>> histSample;
        histSample.push_back(x0);

        // Time evolution
        timeEvolutionRefined[j] = x0;
        if (dynamicsType_==inLongitude)
            timeEvolutionRefined[j].push_back(sGuess[0]);

        // Remove propagated values and Push back in vector for output
        eqX0F[j] = x0;
        if (dynamicsType_==inLongitude)
            eqX0F[j].push_back(sGuess[0]);

        // Integrate for each independent variable interval
        for (auto i = 0; i < ni; i++) {

            double t0 = sGuess[i];
            double tf = sGuess[i+1];
            double ddt = (tf-t0)/((double)nTimeEvolutionDiv);

            for ( auto k = 0 ; k < nTimeEvolutionDiv; k++ )
            {
                double tt0 = t0+((double)k)*ddt;
                double ttf = t0+((double)(k+1))*ddt;

                // Number of integration steps according to size of interval
                auto nIntStep = (int)ceil((double) nint * (ttf-tt0) /
                                          (sGuess.back() - sGuess[0])) + 1;

                // Integrate
                doubleIntegrator.integrate(tt0,ttf, nIntStep, x0, xf);
                x0 = xf;

                // Save for time evolution output
                for (unsigned int w = 0; w < ns; w++)
                    timeEvolutionRefined[j].push_back(xf[w]);
                if (dynamicsType_==inLongitude)
                    timeEvolutionRefined[j].push_back(sGuess[ttf]);
            }

            // Save for output
            for (unsigned int k = 0; k < ns; k++)
                eqX0F[j].push_back(xf[k]);
            if (dynamicsType_==inLongitude)
                eqX0F[j].push_back(sGuess[i+1]);

            // Save history
            histSample.push_back(xf);
        }

        doubleYRefined[j] = constraintParams.pSetCondition(xf);
        doubleFinalSamplesRefined[j] = xf;

        // Save history of sample
        samplesDoubleRefined.push_back(histSample);
    }
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesRefinedDouble.txt", eqX0F);
    eqX0F.clear();
    dlmwrite("../../results/fableApophis/timeEvolutionRefined.txt", timeEvolutionRefined);
    timeEvolutionRefined.clear();








    /**
     * >>>>>>>>>>>>>>>>>>>>>> CHECK POLYNOMIAL PROPAGATION EXAMPLE <<<<<<<<<<<<<<<<<<<<
     */

    cout << endl << ">>>>>>>>>>>>>>>>>>>>>> CHECK POLYNOMIAL PROPAGATION EXAMPLE <<<<<<<<<<<<<<<<<<<<" << endl;


    /**
     * FABLE control
     */

    // Save for output
    vector<vector<vector<double>>> samplesPolynomialFable;
    vector<double> polYFable(nSamples);
    eqX0F = vector<vector<double>>(nSamples);
    carX0F = vector<vector<double>>(nSamples);

    for (auto i = 0; i < nSamples; i++) {
        // Propagate sample [only final state]
        vector<vector<double>> XF = polynomialPropagator.propagateMiddle(eqInitialSamples[i]);

        for (auto j = 0; j <= ni; j++) {

            // Equinoctial vector
            vector<double> eqX(XF[j]), carX(6);
            if (dynamicsType_==inLongitude)
                eqX.push_back(sGuess[j]);

            // Push back in history
            for (double &n: eqX) {
                eqX0F[i].push_back(n);
            }

            // Convert
            smartastro::astrocore::conversion_coordinates::eq2car(eqX,mu,carX);

            // Save in history
            for (auto k = 0 ; k < 3; k++)
                carX0F[i].push_back(carX[k]*L_scale);
            for (auto k = 3 ; k < 6; k++)
                carX0F[i].push_back(carX[k]*L_scale/T_scale);

        }
        polYFable[i] = constraintParams.pSetCondition(XF[ni]);
        polFinalSamplesFable[i] = XF[ni];
        samplesPolynomialFable.push_back(XF);
    }
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesFABLEPoly.txt", eqX0F);        
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesFABLEPolyCartesian.txt", carX0F);
    eqX0F.clear();
    carX0F.clear();


    /**
     * Refined control
     */

    // Set refined control
    propagatorParams.pControlCoefficients = &uRefined;

    // Save for output
    vector<vector<vector<double>>> samplesPolynomialRefined;
    vector<double> polYRefined(nSamples);
    eqX0F = vector<vector<double>>(nSamples);
    carX0F = vector<vector<double>>(nSamples);
    for (auto i = 0; i < nSamples; i++) {
        // Propagate sample [only final state]
        vector<vector<double>> XF = polynomialPropagator.propagateMiddle(eqInitialSamples[i]);

        for (auto j = 0; j <= ni; j++) {

            // Equinoctial vector
            vector<double> eqX(XF[j]), carX(6);
            if (dynamicsType_==inLongitude)
                eqX.push_back(sGuess[j]);

            // Push back in history
            for (double &n: eqX) {
                eqX0F[i].push_back(n);
            }

            // Convert
            smartastro::astrocore::conversion_coordinates::eq2car(eqX,mu,carX);

            // Save in history
            for (auto k = 0 ; k < 3; k++)
                carX0F[i].push_back(carX[k]*L_scale);
            for (auto k = 3 ; k < 6; k++)
                carX0F[i].push_back(carX[k]*L_scale/T_scale);

        }

        polYRefined[i] = constraintParams.pSetCondition(XF[ni]);
        polFinalSamplesRefined[i] = XF[ni];
        samplesPolynomialRefined.push_back(XF);
    }
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesRefinedPoly.txt", eqX0F);
    dlmwrite("../../results/fableApophis/outputPropagatedSamplesRefinedPolyCartesian.txt", carX0F);
    eqX0F.clear();
    carX0F.clear();


//    // Compute RMSE errors
//    vector<vector<double>> RMSEFable(ni + 1, vector<double>(ns, 0.0));
//    vector<vector<double>> RMSERefined(ni + 1, vector<double>(ns, 0.0));
//    for (auto j = 0; j <= ni; j++)
//    {
//        for (unsigned int k = 0; k < ns; k++)
//        {
//            // Compute errors square
//            for (auto i = 0; i < nSamples; i++)
//            {
//                RMSEFable[j][k] += pow(samplesDoubleFable[i][j][k] - samplesPolynomialFable[i][j][k], 2.0);
//                RMSERefined[j][k] += pow(samplesDoubleRefined[i][j][k] - samplesPolynomialRefined[i][j][k], 2.0);
//            }
//
//            // Normalise and take square root
//            RMSEFable[j][k] /= (double) (nSamples);
//            RMSEFable[j][k] = sqrt(RMSEFable[j][k]);
//            RMSERefined[j][k] /= (double) (nSamples);
//            RMSERefined[j][k] = sqrt(RMSERefined[j][k]);
//        }
//    }

//    // Print samples evolution for fable control
//    for (auto i = 0; i < 1; i++)
//    {
//        if (i == 0)
//        {
//            cout << endl << "Fable double vs Poly vs Final error = " << endl;
//            for (auto j = 0; j <= ni; j++)
//            {
//
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesDoubleFable[i][j][k] << " ";
//                if (j==ni)
//                    cout << "; y = " << doubleYFable[i];
//                cout << endl;
//
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesPolynomialFable[i][j][k] << " ";
//                if (j==ni)
//                    cout << "; y = " << polYFable[i];
//
//                cout << endl;
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesDoubleFable[i][j][k] - samplesPolynomialFable[i][j][k] << " ";
//                cout << endl << endl;
//
//            }
//            cout << endl;
//        }
//    }


//    // Print samples evolution for refined control
//    for (auto i = 0; i < 1; i++)
//    {
//        if (i == 0)
//        {
//            cout << endl << "Refined double vs Poly vs Final error = " << endl;
//            for (auto j = 0; j <= ni; j++)
//            {
//
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesDoubleRefined[i][j][k] << " ";
//                if (j==ni)
//                    cout << "; y = " << doubleYRefined[i];
//                cout << endl;
//
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesPolynomialRefined[i][j][k] << " ";
//                if (j==ni)
//                    cout << "; y = " << polYRefined[i];
//                cout << endl;
//
//                for (unsigned int k = 0; k < ns; k++)
//                    cout << samplesDoubleRefined[i][j][k] - samplesPolynomialRefined[i][j][k] << " ";
//                cout << endl << endl;
//
//            }
//            cout << endl;
//        }
//    }


//    cout << endl << "RMSEFable of each state (column) for different times (row) = " << endl;
//    for ( auto j = 0 ; j <=ni; j++ )
//    {
//        cout << "s = " << sGuess[j] << " RMSEFable = " ;
//        cout << RMSEFable[j] << endl;
//    }
//    cout << endl;
//    cout << "RMSERefined of each state (column) for different times (row) = " << endl;
//    for ( auto j = 0 ; j <=ni; j++ )
//    {
//        cout << "s = " << sGuess[j] << " RMSERefined = " ;
//        cout << RMSERefined[j] << endl;
//    }
//    cout << endl;



    double g;
//
    constraintParams.pSamples = &doubleFinalSamplesFable;
    g = finalConstraint.evaluate();
    cout << "Fable double Final Probability = " << g << endl ;

    constraintParams.pSamples = &polFinalSamplesFable;
    g = finalConstraint.evaluate();
    cout << "Fable polynomial Final Probability = " << g << endl ;

    constraintParams.pSamples = &doubleFinalSamplesRefined;
    g = finalConstraint.evaluate();
    cout << "Refined double Final Probability = " << g << endl ;

    constraintParams.pSamples = &polFinalSamplesRefined;
    g = finalConstraint.evaluate();
    cout << "Refined polynomial Final Probability = " << g << endl ;


    cout << "Final Sigma Region: " << endl;
    cout << "LB = ";
    for (auto i = 0 ; i < ns; i++ ){
        cout << finalLB[i] << " ";
    }
    cout << endl;
    cout << "UB = ";
    for (auto i = 0 ; i < ns; i++ ){
        cout << finalUB[i] << " ";
    }
    cout << endl;
    cout << endl;



    return 0;
}
