/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include <iostream>

#include "ipanemaSolvers/ipanemaSolver.h"
#include "ipanemaObjectives/integralOfControl.h"
#include "ipanemaConstraints/samplingProbabilityConstraint.h"
#include "PropagatorSurrogates/intrusivePolynomialPropagator.h"

#include "Dynamics/base_dynamics.h"
#include "Integrators/rk4.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

using namespace std;
using namespace ipanema;
using namespace propagatorSurrogates;
using namespace ipanemaConstraints;
using namespace ipanemaObjectives;
using namespace ipanemaSolvers;

/**
 * Double dynamics
 */
template <class StateType>
class planarFableDynamics : public smartmath::dynamics::base_dynamics<StateType> {

public:

    struct planarFableDynamicsParams {

        // Gravitational parametr
        double                           mu;

        // Time grid
        vector<double>*                  pTimeGrid;

        // Inner vector gathers for control coefficients per each variable
        vector<vector<StateType>>*       pControlCoefficients;

    };


protected:

    const planarFableDynamicsParams* m_pParams;

public:

    planarFableDynamics(const planarFableDynamicsParams* pParams) :
            smartmath::dynamics::base_dynamics<StateType>("planarFableDynamics"),
            m_pParams(pParams)
    {}


    /**
     * Get time index
     */
    unsigned int get_tIndex(const double &t) const
    {

        unsigned int ni = (*m_pParams->pTimeGrid).size()-1;

        double t0, tf;
        for ( unsigned int i = 0 ; i < ni ; i++ )
        {
            t0 = (*m_pParams->pTimeGrid)[i];
            tf = (*m_pParams->pTimeGrid)[i+1];
            if ( t >= t0 && t < tf )
                return i;
        }

        if ( t >= t0 && (t-tf)/(std::min(std::abs(t),std::abs(tf))<1e-10 ))
            return ni-1;

        ipanema_throw("Input time is outside defined time grid: " + "input t = " + std::to_string(t) +
                      "; T grid = [" +
                      std::to_string((*m_pParams->pTimeGrid)[0]) +
                      "," + std::to_string((*m_pParams->pTimeGrid)[ni]) + "]" );
        return 1e5;
    }





    /**
     * @brief evaluate evaluate the dynamics at a given instant of time and a given state.
     *
     * Function to evaluate the dinamics at a given instant of time and a given state. It is a virtual function so any class that inherites from base_dynamics need to implement it.
     * @param[in] t time
     * @param[in] state state values at time t
     * @param[out] dstate derivative of the states at time t
     * @return
     */
    int evaluate(const double &L, const vector<StateType> &x, vector<StateType> &dx) const {

        unsigned int nc = (*m_pParams->pControlCoefficients)[0].size();
        // Sanity check
        if(nc!=2)
            ipanema_throw("2 control variables are mandatory");

        // Time index which indicates discretisation interval
        unsigned tIndex = get_tIndex(L);

        // Control vector [magnitude-angle]
        vector<StateType> u((*m_pParams->pControlCoefficients)[tIndex]) ;

        // Transformed control vector [components]
        vector<StateType> U ;
        U.push_back(u[0] * cos(u[1]));
        U.push_back(u[0] * sin(u[1]));

        // Initialize to u[0]*x[0] polynomial
        dx.clear();

        StateType B2   ( 1.0-pow(x[1],2.0)-pow(x[2],2.0) );
        StateType B4a2 ( pow(1.0-pow(x[1],2.0)-pow(x[2],2.0),2.0)*pow(x[0],2.0) );
        StateType PHI  ( 1.0+x[1]*std::sin(L)+x[2]*std::cos(L) );
        StateType PHI2 ( pow(PHI,2.0) );
        StateType PHI3 ( pow(PHI,3.0) );

        // da/dL
        dx.push_back(pow(x[0], 3.0));
        dx[0] *= 2.0 / m_pParams->mu * B2 * ((x[2] * std::sin(L) - x[1] * std::cos(L)) / PHI2 * U[0] + U[1] / PHI);

        // dP1/dL
        dx.push_back(B4a2);
        dx[1] *= 1.0 / m_pParams->mu *
                 (-std::cos(L) / PHI2 * U[0] + ((x[1] + std::sin(L)) / PHI3 + std::sin(L) / PHI2) * U[1] );

        // dP2/dL
        dx.push_back(B4a2);
        dx[2] *= 1.0 / m_pParams->mu *
                 (std::sin(L) / PHI2 * U[0] + ((x[2] + std::cos(L)) / PHI3 + std::cos(L) / PHI2) * U[1] );

        return 0;

    } // evaluate


};



static const double au = 149597870.691*1.0e3;    // [m]
static const double Tsid = 86164.090517;       // [s]
static const double mu_sun = 1.32712440018e20; // [m3/s2]




int equinoctialPlanar () {


    /**
     * Dimension of the problem
     */

    unsigned int ns       = 3;      // Number of state variables
    unsigned int nc       = 2;      // Number of control variables
    unsigned int ng       = 1;      // Number of constraints
    unsigned int ni;                // Number of discretisation intervals (input)
    unsigned int nint     = 2000;   // Number of integration steps for the whole interval
    unsigned int nSamples = 200;    // Number of integration steps for the whole interval

    // Polynomial settings
    int nVar = ns+nc;
    int nDeg = 6;
    bool monomial = true;


    // Scales
    double L_scale = au;
    double T_scale = 1.0e8;
    vector<double> controlScale(nc,1.0);
    vector<double> constraintScale (ng,1.0);
    double objectiveScale = 10.0;

    /**
     * Fixed quantity
     */

    double maxT = 0.053 / L_scale * pow(T_scale,2.0) ; // Maximum thrust FABLE [N=kg*m/s^2]
    double m0 = 644.3; // Initial mass [kg]
    unsigned int nRev = 1;
    double mu = mu_sun / pow(L_scale,3.0) * pow(T_scale,2.0); // Sun gravity constant [m^3/(s^2)]


    /**
     * Initial and final boundary conditions
     */

    vector<double> refInitialCondition (ns);
    vector<double> refFinalCondition (ns);
//
//    refInitialCondition[0] = 1.0; // a [au]
//    refInitialCondition[1] = 0.0; // p1 [rad]
//    refInitialCondition[2] = 0.0; // p2 [rad]
//
//    refFinalCondition[0] = 1.1; // a [au]
//    refFinalCondition[1] = 0.0; // p1 [rad]
//    refFinalCondition[2] = 0.0; // p2 [rad]


    /**
     * Read input guesses
     */

    // Read input file by FABLE
    string inputFile ("planarLowThrust/Cristian_eqSimple.txt");
    vector<vector<double>> inputFABLE;
    dlmread(inputFile,inputFABLE);
    unsigned int inputD = inputFABLE.size();


    /**
     * Time discretization
     */

    // Number of control discretization intervals
    ni = inputD-1;

    // Time vector
    vector<double> t_guess (ni+1) ;
    for ( unsigned int i = 0 ; i < inputD; i++ )
        t_guess[i] = inputFABLE[i][0] / T_scale ; // Scale time originally in s

    /**
     * L [independent variable] discretization
     */

    // L vector
    vector<double> L_guess (ni+1) ;
    for ( unsigned int i = 0 ; i < inputD; i++ )
        L_guess[i] = inputFABLE[i][6];

    /**
     * State guess
     */
    vector<vector<double>> x_guess (ni+1,vector<double>(ns));
    for ( unsigned int i = 0 ; i < inputD; i++ )
    {
        for (unsigned int j = 0; j < ns; j++)
        {
            x_guess[i][j] = inputFABLE[i][j+1];
        }

        // Scale semi-major axis originally in m
        x_guess[i][0] /= L_scale;
    }


    /**
     * Control parameter guess
     */
    vector<vector<double>> u_guess ( ni,
                                     vector<double>( nc,0.0)
    );

    for ( unsigned int i = 0 ; i < ni; i++ )
    {
        for (unsigned int j = 0; j < nc; j++)
        {
            u_guess[i][j] = inputFABLE[i][j+7];
        }

        u_guess[i][0] *= pow(T_scale,2.0)/L_scale; // epsilon - originally [m/s^2]
    }




    /**
     * Print guesses and refined control
     */
    cout << "Time, State and Control guess and Control refined:" << endl;
    for (unsigned int i = 0; i <= ni; i++) {
        cout << t_guess[i] << " ";

        for (unsigned int j = 0; j < ns; j++)
            cout << x_guess[i][j] << " ";

        cout << L_guess[i] << " ";

        if (i != ni) {
            for (unsigned int j = 0; j < nc; j++)
                cout << u_guess[i][j] << " ";
        }
        cout << endl;
    }
    cout << endl;

//    return 0;



    /**
     * Control parameter bounds [simplified]
     */

    vector<vector<double>> realUBounds(nc,vector<double>(2));
    double deltaControl = 0.0;

    // Magnitude
    realUBounds[0][0] = 0.0-deltaControl;
    realUBounds[0][1] = maxT/m0+1.0e-3+deltaControl;
    // Angle
    realUBounds[1][0] = 0.17;
    realUBounds[1][1] = 3.0;

    // Dynamics real bounds (cause worhp checks slightly more)
    vector<vector<double>> largerUBounds(realUBounds);
    for (unsigned int i = 0 ; i < nc; i++)
    {
        largerUBounds[i][0] -= 1.001e-5;
        largerUBounds[i][1] += 1.001e-5;
    }


    /**
     * Initial starting box zone
     */

    // Define relative to reference value magnitude
    vector<double> initialMean(refInitialCondition),
            initialSigma (ns),
            initialLB(ns),
            initialUB(ns);


//    // Symmetric Initial Box
//    initialLB[0] = initialMean[0]-initialSigma[0];
//    initialLB[1] = initialMean[1]-initialSigma[1];
//    initialLB[2] = initialMean[2]-initialSigma[2];
//
//    initialUB[0] = initialMean[0]+initialSigma[0];
//    initialUB[1] = initialMean[1]+initialSigma[1];
//    initialUB[2] = initialMean[2]+initialSigma[2];

    // Asymmetric Initial Box interval
//    initialLB[0] = initialMean[0];
//    initialLB[1] = initialMean[1]-initialSigma[1];
//    initialLB[2] = initialMean[2]-initialSigma[2];
//
//    initialUB[0] = initialMean[0]+initialSigma[0];
//    initialUB[1] = initialMean[1]+initialSigma[1];
//    initialUB[2] = initialMean[2]+initialSigma[2];

    // Custom
    initialLB[0] = +0.995 ;
    initialLB[1] = -0.01;
    initialLB[2] = -0.006;

    initialUB[0] = +1.000001;
    initialUB[1] = +0.01;
    initialUB[2] = +0.000001;

    // Define symmetric sigma
    double sigmaMult = 4.0;
    for ( unsigned int i = 0 ; i < ns ; i++ )
    {
        initialSigma[i] = (initialUB[i]-initialLB[i])/sigmaMult;
    }





    /**
     * Define final target ellipsoid dimension
     */

    // Final mean and sigma
    vector<double> finalMean (refFinalCondition),
            finalSigma(ns),
            finalLB(ns),
            finalUB(ns);

    // Define relative to reference value magnitude
//    finalSigma[0] = 5.0e-3/2.0; // First is au quite big
//    finalSigma[1] = 1.0e-2; // Radians
//    finalSigma[2] = 6.0e-3/2.0; // Radians

    // Define relative to reference value magnitude
    finalSigma[0] = 6.0e-3/2.0; // First is au quite big
    finalSigma[1] = 1.1e-2; // Radians
    finalSigma[2] = 7.0e-3/2.0; // Radians

    //    // Symmetric Initial Box
    finalLB[0] = finalMean[0]-finalSigma[0];
    finalLB[1] = finalMean[1]-finalSigma[1];
    finalLB[2] = finalMean[2]-finalSigma[2];

    finalUB[0] = finalMean[0]+finalSigma[0];
    finalUB[1] = finalMean[1]+finalSigma[1];
    finalUB[2] = finalMean[2]+finalSigma[2];

    // Asymmetric Initial Box interval
//    finalLB[0] = finalMean[0];
//    finalLB[1] = finalMean[1]-finalSigma[1];
//    finalLB[2] = finalMean[2]-finalSigma[2];
//
//    finalUB[0] = finalMean[0]+finalSigma[0];
//    finalUB[1] = finalMean[1]+finalSigma[1];
//    finalUB[2] = finalMean[2]+finalSigma[2];




    // Print output mean, LB, UB for plot
    vector<vector<double>> initialFinalMeanLBUBOutput (2,vector<double>(3*ns));
    // Initial mean - LB - UB
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i]      = initialMean[i];
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i+ns]   = initialLB[i];
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[0][i+2*ns] = initialUB[i];
    // Final mean - LB - UB
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i]      = finalMean[i];
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i+ns]   = finalLB[i];
    for (unsigned int i = 0; i < ns; i++)
        initialFinalMeanLBUBOutput[1][i+2*ns] = finalUB[i];
    dlmwrite("../../results/planarLowThrust/initialFinalMeanLBUB.txt",initialFinalMeanLBUBOutput);
    initialFinalMeanLBUBOutput.clear();





    /**
     * Initial polynomial box
     */

    vector<smartuq::polynomial::chebyshev_polynomial> initialPolynomials;
    for (unsigned int i = 0 ; i < ns; i++ )
    {
        initialPolynomials.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar,nDeg,i,
                                                          initialLB[i],
                                                          initialUB[i],
                                                          monomial)
        );
    }



    /**
     * Polynomial control
     */

    // The same for all the times
    vector<smartuq::polynomial::chebyshev_polynomial> polynomialControl_i;
    for ( int i = 0 ; i < nc; i++ )
    {
        polynomialControl_i.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar,nDeg,i+ns,
                                                          largerUBounds[i][0],
                                                          largerUBounds[i][1],
                                                          monomial)
        );
    }

    // Same for all discretisation interval
    vector<vector<smartuq::polynomial::chebyshev_polynomial>> polynomialControl (ni,polynomialControl_i);



    /**
     *  Lambda function to compute final ellipsoid condition
     */

    function<double(std::vector<double>)> ellipsoidSampleSetCondition =
            [mean=finalMean,sigma=finalSigma](const std::vector<double>& x) -> double
            {
                unsigned int ns = x.size();
                double y = 0.0;

                // Evaluate ellipsoid contour
                for ( unsigned int j = 0 ; j < ns; j++ ) {
                    y += std::pow((x[j]-mean[j])/sigma[j], 2.0);
                }

                return y;
            };



    /**
     * Quantities to be optimised
     * (initialised to guess)
     */
    vector<double> L_refined (L_guess) ;
    vector<vector<double>> u_refined(u_guess);


    /**
     * Initialise dynamics and integrator
     */

    // Dynamics parameters
    planarFableDynamics<smartuq::polynomial::chebyshev_polynomial>::planarFableDynamicsParams dynamicsParams;
    dynamicsParams.mu                   = mu;
    dynamicsParams.pControlCoefficients = &polynomialControl;
    dynamicsParams.pTimeGrid            = &L_refined;

    // Dynamics
    planarFableDynamics<smartuq::polynomial::chebyshev_polynomial>* polynomialDynamics =
            new planarFableDynamics<smartuq::polynomial::chebyshev_polynomial>(&dynamicsParams);

    // Integrator
    smartmath::integrator::rk4<smartuq::polynomial::chebyshev_polynomial> polynomialIntegrator (polynomialDynamics);



    /**
     * Initial and final samples
     */

    vector<vector<double>> initialSamples(nSamples+1,vector<double>(ns));
    vector<vector<double>> finalSamples(nSamples+1,vector<double>(ns));


    initialSamples[0] = initialMean;

    // Generate initial samples
    gsl_rng * r = gsl_rng_alloc(gsl_rng_taus);
    gsl_rng_set(r,11);

    // Second component p1 [rad] according to (manually) truncated Gaussian distribution
    for (unsigned int i = 0, cSamples = 0 ; i < ns; i++)
    {
        cSamples = 0;
        while (cSamples<nSamples)
        {
            // Draw a sample
            double sample = gsl_ran_gaussian(r,initialSigma[i]) + initialMean[i];

            // If within range, set sample and update counter
            if (sample>=initialLB[i] && sample<=initialUB[i])
            {
                initialSamples[1+cSamples++][i] = sample;
            }
        }
    }

    // Free r
    gsl_rng_free (r);

//    // Print initial samples
//    cout << "Initial samples: " << endl;
//    for (unsigned int i = 0 ; i < nSamples; i++)
//    {
//        for (unsigned int j = 0 ; j < ns; j++)
//        {
//            cout << initialSamples[i][j] << " ";
//        }
//        cout << endl;
//    }
//
//    cout << endl << "InitialLB = ";
//    for (unsigned int j = 0 ; j < ns; j++)
//    {
//        cout << initialLB[j] << " ";
//    }
//    cout << endl << "InitialUB = ";
//    for (unsigned int j = 0 ; j < ns; j++)
//    {
//        cout << initialUB[j] << " ";
//    }
//    cout << endl;
//
//    return 0;



    /**
     * Auxiliary quantities to be shared through pointer in several classes
     */
    bool majorIteratior = false; // Belongs to constraint and NLP solver


    /** 
     * Surrogate propagator
     */

    // Parameters
    intrusivePolynomialPropagator::intrusivePolynomialPropagatorParams propagatorParams;

    propagatorParams.pStartPolynomials    = &initialPolynomials;
    propagatorParams.pIntegrator          = &polynomialIntegrator;
    propagatorParams.pTimeGrid            = &L_refined;
    propagatorParams.pControlCoefficients = &u_refined;
    propagatorParams.controlBounds        = largerUBounds;
    propagatorParams.nIntegrationSteps    = nint;
    propagatorParams.printToFile          = true;
    propagatorParams.outputFilePrefix     = "../../results/planarLowThrust/";
    propagatorParams.readFromFile         = true;
    propagatorParams.inputFilePrefix      = propagatorParams.outputFilePrefix;

    // Predefined range for polynomial propagation
    propagatorParams.usePredefinedRange   = true;
    propagatorParams.predefinedRange      = vector<vector<double>>(ns,vector<double>(2,0.0));
    // a [au]
    propagatorParams.predefinedRange[0][0] = +0.97;
    propagatorParams.predefinedRange[0][1] = +1.15;
    // p1 [rad]
    propagatorParams.predefinedRange[1][0] = -0.04;
    propagatorParams.predefinedRange[1][1] = +0.05;
    // p2 [rad]
    propagatorParams.predefinedRange[2][0] = -0.05;
    propagatorParams.predefinedRange[2][1] = 0.03;

    // Initialise surrogate propagator
    intrusivePolynomialPropagator polynomialPropagator(&propagatorParams);

    // Explicitly build polynomial surrogate
    polynomialPropagator.build();

    



    /**
     * Final probabilistic constraint
     */

    vector<double> constraintWorhpBounds(2);
    constraintWorhpBounds[0] = -1.0e21;
    constraintWorhpBounds[1] = 5.0e-2;

    // To judge wheter inside target zone or outside
    double constraintSetLB = 0.0;
    double constraintSetUB = 1.0;

    // Parameters
    samplingProbabilityConstraint::samplingProbabilityConstraintParams constraintParams;

    constraintParams.pSamples           = &finalSamples;
    constraintParams.sampleSubsetOption = false;
    constraintParams.expectationMC      = true; // Sampled according to initial distribution
    constraintParams.pUseFullSampleSet  = &majorIteratior;
    constraintParams.membershipType     = samplingProbabilityConstraint::BIQUADRATIC_MEMBERSHIP;
    constraintParams.setValueLB         = constraintSetLB;
    constraintParams.setValueUB         = constraintSetUB;
    constraintParams.pSetCondition      = ellipsoidSampleSetCondition;
    
    // Initialise constraint
    samplingProbabilityConstraint finalConstraint(&constraintParams);


    

    /**
     * Objective class
     */

    // Params
    integralOfControl::integralOfControlParams objectiveParams;

    objectiveParams.pTimeGrid              = &L_refined;
    objectiveParams.pControlCoefficients   = &u_refined;
    objectiveParams.objectiveIntegrandType = integralOfControl::MODULUS_OBJECTIVE;
    objectiveParams.controlType            = integralOfControl::MAGNITUDEANGLES_CONTROL;

    // Initialise objective
    integralOfControl objective(&objectiveParams);

    

    /**
     * Ipanema solver
     */

    // Parameters
    ipanemaSolver::ipanemaSolverParams solverParams;

    solverParams.pControlCoefficients   = &u_refined;
    solverParams.pTimeGrid              = &L_refined;
    solverParams.pInitialSamples        = &initialSamples;
    solverParams.pFinalSamples          = &finalSamples;
    solverParams.pPropagator            = bind(&intrusivePolynomialPropagator::propagate,
                                               &polynomialPropagator,
                                               std::placeholders::_1);
    solverParams.pObjective             = bind(&integralOfControl::evaluate,
                                               &objective);
    solverParams.pConstraints           = vector<function<double()>>(1,bind(&samplingProbabilityConstraint::evaluate,
                                                                            &finalConstraint));
    solverParams.constraintsBounds      = vector<vector<double>>(ng,constraintWorhpBounds);
    solverParams.controlBounds          = realUBounds;
    solverParams.constraintScale        = constraintScale;
    solverParams.controlScale           = controlScale;
    solverParams.pMajorIteration        = &majorIteratior;
    solverParams.objectiveScale         = objectiveScale;
    solverParams.constraintRelax        = true;
    solverParams.gRelaxPenalty          = vector<double>(ng,1.0);
    solverParams.gRelaxUB               = 1.5e-2;

    // Initialise solver
    ipanemaSolver solver(&solverParams);

    




    /**
     * Print ms dimension
     */

    cout << "Number of multiple shooting states = " << ns << endl;
    cout << "Number of multiple shooting controls = " << nc << endl;
    cout << "Number of multiple shooting intervals = " << ni << endl;
    cout << "Number of ipanema polynomial variables = " << nVar << endl;
    cout << "Degree of ipanema polynomial = " << nDeg << endl;


    /**
     * Solve by WORHP
     */

    solver.solveByWorhp();


    /**
     * Print final quantities
     */

    // Final free vector
    vector<double> x_refined = solver.get_x();
    vector<vector<double>> xFreeVariables (1,x_refined);

    // Save control
    vector<vector<double>> outputControl (ni,vector<double>(nc));
    vector<vector<double>> guessControl (ni,vector<double>(nc));
    for ( unsigned int i = 0, cc = 0 ; i < ni; i++ )
    {
        for ( unsigned int j = 0 ; j < nc; j++ )
        {
            outputControl[i][j] = u_refined[i][j];
            guessControl[i][j]  = u_guess[i][j];
        }
    }

    cout << "Final free vector: " ;
    for ( unsigned int i = 0 ; i < x_refined.size(); i++ )
        cout << x_refined[i] << " " ;
    cout << endl;

    dlmwrite("../../results/planarLowThrust/refinedControlDeg" +
             std::to_string(nDeg) +".txt",outputControl);
    dlmwrite("../../results/planarLowThrust/fableControlDeg" +
             std::to_string(nDeg) +".txt",guessControl);

    double finalObjControl = solver.get_J();
    if(solverParams.constraintRelax)
        finalObjControl -= solverParams.gRelaxPenalty[0]*x_refined.back();
    cout << "Final control objective unscaled  = " << finalObjControl << endl;

}




int reintegrateEquinoctialPlanar () {


    /**
     * Dimension of the problem
     */

    unsigned int ns = 3;      // Number of state variables
    unsigned int nc = 2;      // Number of control variables
    unsigned int ng = 1;      // Number of constraints
    unsigned int ni;                // Number of discretisation intervals (input)
    unsigned int nint = 2000;   // Number of integration steps for the whole interval
    unsigned int nSamples = 500;    // Number of integration steps for the whole interval

    // Polynomial settings
    int nVar = ns + nc;
    int nDeg = 6;
    bool monomial = true;


    // Scales
    double L_scale = au;
    double T_scale = 1.0e8;
    vector<double> controlScale(nc, 1.0);
    vector<double> constraintScale(ng, 1.0);
    double objectiveScale = 1.0;

    /**
     * Fixed quantity
     */

    double maxT = 0.053 / L_scale * pow(T_scale, 2.0); // Maximum thrust FABLE [N=kg*m/s^2]
    double m0 = 644.3; // Initial mass [kg]
    unsigned int nRev = 1;
    double mu = mu_sun / pow(L_scale, 3.0) * pow(T_scale, 2.0); // Sun gravity constant [m^3/(s^2)]


    /**
     * Initial and final boundary conditions
     */

    vector<double> refInitialCondition(ns);
    vector<double> refFinalCondition(ns);

    refInitialCondition[0] = 1.0; // a [au]
    refInitialCondition[1] = 0.0; // p1 [rad]
    refInitialCondition[2] = 0.0; // p2 [rad]

    refFinalCondition[0] = 1.1; // a [au]
    refFinalCondition[1] = 0.0; // p1 [rad]
    refFinalCondition[2] = 0.0; // p2 [rad]


    /**
     * Read input guesses
     */

    // Read input file by FABLE
    string inputFile("planarLowThrust/Cristian_eq.txt");
    vector<vector<double>> inputFABLE;
    dlmread(inputFile, inputFABLE);
    unsigned int inputD = inputFABLE.size();


    /**
     * Time discretization
     */

    // Number of control discretization intervals
    ni = inputD - 1;

    // Time vector
    vector<double> t_guess(ni + 1);
    for (unsigned int i = 0; i < inputD; i++)
        t_guess[i] = inputFABLE[i][0] / T_scale; // Scale time originally in s

    /**
     * L [independent variable] discretization
     */

    // L vector
    vector<double> L_guess(ni + 1);
    for (unsigned int i = 0; i < inputD; i++)
        L_guess[i] = inputFABLE[i][6];

    /**
     * State guess
     */
    vector<vector<double>> x_guess(ni + 1, vector<double>(ns));
    for (unsigned int i = 0; i < inputD; i++) {
        for (unsigned int j = 0; j < ns; j++) {
            x_guess[i][j] = inputFABLE[i][j + 1];
        }

        // Scale semi-major axis originally in m
        x_guess[i][0] /= L_scale;
    }


    /**
     * Control parameter guess
     */
//    vector<double> uHardCoded {0.000202506, 2.0, 5.48072, 2.72973, 5.30544, 1.55979, 0.000202506, 1.55979, 5.34307, 1.57615, 5.48116, 1.70317, 0.000202506, 1.70317};
    vector<vector<double>> u_guess(ni,
                                   vector<double>(nc, 0.0)
    );

    for (unsigned int i = 0; i < ni; i++) {
        for (unsigned int j = 0; j < nc; j++) {
            u_guess[i][j] = inputFABLE[i][j + 7];
        }

        u_guess[i][0] *= pow(T_scale, 2.0) / L_scale; // epsilon - originally [m/s^2]
    }
//    for ( unsigned int i = 0, cu = 0 ; i < ni; i++ )
//    {
//        for (unsigned int j = 0; j < nc; j++)
//        {
//            u_guess[i][j] = uHardCoded[cu++];
//        }
//    }


    /**
     * Refined control
     */

    vector<vector<double>> u_refined(u_guess);
    dlmread("../../results/planarLowThrust/refinedControlDeg" + std::to_string(nDeg) + ".txt", u_refined);



    /**
     * Initial/Final mean and sigma
     */

    vector<vector<double>> initialFinalMeanLBUBInput;
    dlmread("../../results/planarLowThrust/initialFinalMeanLBUB.txt", initialFinalMeanLBUBInput);

    vector<double> initialMean(ns), finalMean(ns),
            initialLB(ns), initialUB(ns),
            finalLB(ns), finalUB(ns);

    for (unsigned int i = 0; i < ns; i++) {
        // Mean
        initialMean[i] = initialFinalMeanLBUBInput[0][i];
        finalMean[i] = initialFinalMeanLBUBInput[1][i];

        // LB
        initialLB[i] = initialFinalMeanLBUBInput[0][i + ns];
        finalLB[i] = initialFinalMeanLBUBInput[1][i + ns];

        // UB
        initialUB[i] = initialFinalMeanLBUBInput[0][i + 2 * ns];
        finalUB[i] = initialFinalMeanLBUBInput[1][i + 2 * ns];
    }

    vector<double> initialSigma(ns),
            finalSigma(ns);
    double sigmaMult = 4.0;
    for (unsigned int i = 0; i < ns; i++) {
        initialSigma[i] = (initialUB[i] - initialLB[i]) / sigmaMult;
        finalSigma[i] = (finalUB[i] - finalLB[i]) / 2.0;
    }

//    // Define relative to reference value magnitude
//    initialSigma[0] = std::abs(1.0*(1.0e-3)*refInitialCondition[0]); // First is au quite big
//    for ( unsigned int i = 1 ; i < ns; i++ )
//        initialSigma[i] = std::abs(1.0e-2); // Radians
//
//    // Define relative to reference value magnitude
//    finalSigma[0] = std::abs(5.0 * (1.0e-3) * finalMean[0]); // First is au quite big
//    for (unsigned int i = 1; i < ns; i++)
//        finalSigma[i] = std::abs(5.0 * (1.0e-2)); // Radians


    /**
     * Control parameter bounds [simplified]
     */

    vector<vector<double>> realUBounds(nc, vector<double>(2));
    double deltaControl = 0.0;

    // Magnitude
    realUBounds[0][0] = 0.0 - deltaControl;
    realUBounds[0][1] = maxT / m0 + 1.0e-3 + deltaControl;
    // Angle
    realUBounds[1][0] = 0.17;
    realUBounds[1][1] = 3.0;
    // Angle [simplified]
//    realUBounds[1][0] = 1.0;
//    realUBounds[1][1] = 3.0;


    // Dynamics real bounds (cause worhp checks slightly more)
    vector<vector<double>> largerUBounds(realUBounds);
    for (unsigned int i = 0; i < nc; i++) {
        largerUBounds[i][0] -= 1.001e-5;
        largerUBounds[i][1] += 1.001e-5;
    }


    /**
     * Initial polynomial box
     */

    vector<smartuq::polynomial::chebyshev_polynomial> initialPolynomials;
    for (unsigned int i = 0; i < ns; i++) {
        initialPolynomials.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar, nDeg, i,
                                                          initialLB[i],
                                                          initialUB[i],
                                                          monomial)
        );
    }


    /**
     * Polynomial control
     */

    // The same for all the times
    vector<smartuq::polynomial::chebyshev_polynomial> polynomialControl_i;
    for (int i = 0; i < nc; i++) {
        polynomialControl_i.push_back(
                smartuq::polynomial::chebyshev_polynomial(nVar, nDeg, i + ns,
                                                          largerUBounds[i][0],
                                                          largerUBounds[i][1],
                                                          monomial)
        );
    }

    // Same for all discretisation interval
    vector<vector<smartuq::polynomial::chebyshev_polynomial>> polynomialControl(ni, polynomialControl_i);


    /**
     *  Lambda function to compute final ellipsoid condition
     */

    function<double(std::vector<double>)> ellipsoidSampleSetCondition =
            [mean = finalMean, sigma = finalSigma](const std::vector<double> &x) -> double {
                unsigned int ns = x.size();
                double y = 0.0;

                // Evaluate ellipsoid contour
                for (unsigned int j = 0; j < ns; j++) {
                    y += std::pow((x[j] - mean[j]) / sigma[j], 2.0);
                }

                return y;
            };



    /**
     * Final probabilistic constraint
     */

    vector<double> constraintWorhpBounds(2);
    constraintWorhpBounds[0] = -1.0e21;
    constraintWorhpBounds[1] = 1.0e-6;

    // To judge wheter inside target zone or outside
    double constraintSetLB = 0.0;
    double constraintSetUB = 1.0;

    // Parameters
    samplingProbabilityConstraint::samplingProbabilityConstraintParams constraintParams;

    constraintParams.sampleSubsetOption = false;
    constraintParams.expectationMC = true; // Sampled according to initial distribution
    constraintParams.membershipType = samplingProbabilityConstraint::BIQUADRATIC_MEMBERSHIP;
    constraintParams.setValueLB = constraintSetLB;
    constraintParams.setValueUB = constraintSetUB;
    constraintParams.pSetCondition = ellipsoidSampleSetCondition;

    // Initialise constraint
    samplingProbabilityConstraint finalConstraint(&constraintParams);




    /**
     * Initialise dynamics and integrator
     */

    // Dynamics parameters
    planarFableDynamics<smartuq::polynomial::chebyshev_polynomial>::planarFableDynamicsParams polDynamicsParams;
    polDynamicsParams.mu = mu;
    polDynamicsParams.pControlCoefficients = &polynomialControl;
    polDynamicsParams.pTimeGrid = &L_guess;

    // Dynamics
    planarFableDynamics<smartuq::polynomial::chebyshev_polynomial> *polynomialDynamics =
            new planarFableDynamics<smartuq::polynomial::chebyshev_polynomial>(&polDynamicsParams);

    // Integrator
    smartmath::integrator::rk4<smartuq::polynomial::chebyshev_polynomial> polynomialIntegrator(polynomialDynamics);





    /**
     * Surrogate propagator
     */

    // Parameters
    intrusivePolynomialPropagator::intrusivePolynomialPropagatorParams propagatorParams;

    propagatorParams.pStartPolynomials = &initialPolynomials;
    propagatorParams.pIntegrator = &polynomialIntegrator;
    propagatorParams.pTimeGrid = &L_guess;
    propagatorParams.pControlCoefficients = &u_guess;
    propagatorParams.controlBounds = largerUBounds;
    propagatorParams.nIntegrationSteps = nint;
    propagatorParams.printToFile = true;
    propagatorParams.outputFilePrefix = "../../results/planarLowThrust/";
    propagatorParams.readFromFile = true;
    propagatorParams.inputFilePrefix = propagatorParams.outputFilePrefix;


    // Predefined range for polynomial propagation
    propagatorParams.usePredefinedRange = true;
    propagatorParams.predefinedRange = vector<vector<double>>(ns, vector<double>(2, 0.0));
    // a [au]
    propagatorParams.predefinedRange[0][0] = +0.97;
    propagatorParams.predefinedRange[0][1] = +1.15;
    // p1 [rad]
    propagatorParams.predefinedRange[1][0] = -0.04;
    propagatorParams.predefinedRange[1][1] = +0.05;
    // p2 [rad]
    propagatorParams.predefinedRange[2][0] = -0.05;
    propagatorParams.predefinedRange[2][1] = 0.03;

    // Initialise surrogate propagator
    intrusivePolynomialPropagator polynomialPropagator(&propagatorParams);

    // Explicitly build polynomial surrogate
    polynomialPropagator.build();



    /**
     * Set of samples to propagate for check
     */

    vector<vector<double>> initialSamples(nSamples + 1,
                                          vector<double>(ns));
    vector<vector<double>> doubleFinalSamplesFable(nSamples + 1,
                                                   vector<double>(ns));
    vector<vector<double>> polFinalSamplesFable(nSamples + 1,
                                                vector<double>(ns));
    vector<vector<double>> doubleFinalSamplesRefined(nSamples + 1,
                                                     vector<double>(ns));
    vector<vector<double>> polFinalSamplesRefined(nSamples + 1,
                                                  vector<double>(ns));


    initialSamples[0] = initialMean;

    // Generate initial samples
    gsl_rng *r = gsl_rng_alloc(gsl_rng_taus);
    gsl_rng_set(r, 11);

    // Second component p1 [rad] according to (manually) truncated Gaussian distribution
    for (unsigned int i = 0, cSamples = 0; i < ns; i++) {
        cSamples = 0;
        while (cSamples < nSamples) {
            // Draw a sample
            double sample = gsl_ran_gaussian(r, initialSigma[i]) + initialMean[i];

            // If within range, set sample and update counter
            if (sample >= initialLB[i] && sample <= initialUB[i]) {
                initialSamples[1 + cSamples++][i] = sample;
            }
        }
    }

    // Free r
    gsl_rng_free(r);



    /**
     * Double Numerical Propagation
     */

    cout << endl << ">>>>>>>>>> Check double propagation <<<<<<<<<<" << endl;



    /**
     * Double dynamics and integrator
     */

    // Dynamics parameters
    planarFableDynamics<double>::planarFableDynamicsParams dynamicsParams;
    dynamicsParams.mu = mu;
    dynamicsParams.pControlCoefficients = &u_guess;
    dynamicsParams.pTimeGrid = &L_guess;

    // Dynamics
    planarFableDynamics<double> *doubleDynamics =
            new planarFableDynamics<double>(&dynamicsParams);

    // Integrator
    smartmath::integrator::rk4<double> doubleIntegrator(doubleDynamics);

    /**
     * Propagate for FABLE control
     */

    // For fable solution
    vector<vector<vector<double>>> samplesDoubleFable;
    vector<double> doubleYFable(nSamples + 1);
    vector<vector<double>> X0F(nSamples + 1);

    vector<double> x0(ns), xf(ns);
    for (unsigned int j = 0; j <= nSamples; j++) {

        // Initialise state to
        x0 = initialSamples[j];

        // History of each sample
        vector<vector<double>> histSample;
        histSample.push_back(x0);

        // Remove propagated values and Push back in vector for output
        X0F[j] = x0;
        X0F[j].push_back(L_guess[0]);

        // Integrate for each independent variable interval
        for (unsigned int i = 0; i < ni; i++) {

            // Number of integration steps according to size of interval
            unsigned int nIntStep = std::ceil((double) nint * (L_guess[i + 1] - L_guess[i]) /
                                              (L_guess.back() - L_guess[0])) + 1;

            // Integrate
            doubleIntegrator.integrate(L_guess[i], L_guess[i + 1], nIntStep, x0, xf);
            x0 = xf;

            // Save for output
            for (unsigned int k = 0; k < ns; k++)
                X0F[j].push_back(xf[k]);
            X0F[j].push_back(L_guess[i + 1]);

            // Save history
            histSample.push_back(xf);
        }

        doubleYFable[j] = ellipsoidSampleSetCondition(xf);
        doubleFinalSamplesFable[j] = xf;

        // Save history of sample
        samplesDoubleFable.push_back(histSample);
    }
    dlmwrite("../../results/planarLowThrust/outputPropagatedSamplesFABLEDouble.txt", X0F);
    X0F.clear();

    /**
     * Propagate for refined control
     */

    // Set control parameters
    dynamicsParams.pControlCoefficients = &u_refined;

    // For fable solution
    vector<vector<vector<double>>> samplesDoubleRefined;
    vector<double> doubleYRefined(nSamples + 1);
    X0F = vector<vector<double>>(nSamples + 1);

    for (unsigned int j = 0; j <= nSamples; j++) {

        // Initialise state to
        x0 = initialSamples[j];

        // History of each sample
        vector<vector<double>> histSample;
        histSample.push_back(x0);

        // Remove propagated values and Push back in vector for output
        X0F[j] = x0;
        X0F[j].push_back(L_guess[0]);

        // Integrate for each independent variable interval
        for (unsigned int i = 0; i < ni; i++) {

            // Number of integration steps according to size of interval
            unsigned int nIntStep = std::ceil((double) nint * (L_guess[i + 1] - L_guess[i]) /
                                              (L_guess.back() - L_guess[0])) + 1;

            // Integrate
            doubleIntegrator.integrate(L_guess[i], L_guess[i + 1], nIntStep, x0, xf);
            x0 = xf;

            // Save for output
            for (unsigned int k = 0; k < ns; k++)
                X0F[j].push_back(xf[k]);
            X0F[j].push_back(L_guess[i + 1]);

            // Save history
            histSample.push_back(xf);
        }

        doubleYRefined[j] = ellipsoidSampleSetCondition(xf);
        doubleFinalSamplesRefined[j] = xf;

        // Save history of sample
        samplesDoubleRefined.push_back(histSample);
    }
    dlmwrite("../../results/planarLowThrust/outputPropagatedSamplesRefinedDouble.txt", X0F);
    X0F.clear();








    /**
     * >>>>>>>>>>>>>>>>>>>>>> CHECK POLYNOMIAL PROPAGATION EXAMPLE <<<<<<<<<<<<<<<<<<<<
     */

    cout << endl << ">>>>>>>>>>>>>>>>>>>>>> CHECK POLYNOMIAL PROPAGATION EXAMPLE <<<<<<<<<<<<<<<<<<<<" << endl;


    /**
     * FABLE control
     */

    // Save for output
    vector<vector<vector<double>>> samplesPolynomialFable;
    vector<double> polYFable(nSamples + 1);
    X0F = vector<vector<double>>(nSamples + 1);
    for (unsigned int i = 0; i <= nSamples; i++) {
        // Propagate sample [only final state]
        vector<vector<double>> XF = polynomialPropagator.propagateMiddle(initialSamples[i]);

        for (unsigned int j = 0; j <= ni; j++) {
            for (unsigned int k = 0; k < ns; k++) {
                X0F[i].push_back(XF[j][k]);
            }
            X0F[i].push_back(L_guess[j]);
        }
        polYFable[i] = ellipsoidSampleSetCondition(XF[ni]);
        polFinalSamplesFable[i] = XF[ni];
        samplesPolynomialFable.push_back(XF);
    }
    dlmwrite("../../results/planarLowThrust/outputPropagatedSamplesFABLEPoly.txt", X0F);
    X0F.clear();


    /**
     * Refined control
     */

    // Set refined control
    propagatorParams.pControlCoefficients = &u_refined;

    // Save for output
    vector<vector<vector<double>>> samplesPolynomialRefined;
    vector<double> polYRefined(nSamples + 1);
    X0F = vector<vector<double>>(nSamples + 1);
    for (unsigned int i = 0; i <= nSamples; i++) {
        // Propagate sample [only final state]
        vector<vector<double>> XF = polynomialPropagator.propagateMiddle(initialSamples[i]);

        for (unsigned int j = 0; j <= ni; j++) {
            for (unsigned int k = 0; k < ns; k++) {
                X0F[i].push_back(XF[j][k]);
            }
            X0F[i].push_back(L_guess[j]);
        }
        polYRefined[i] = ellipsoidSampleSetCondition(XF[ni]);
        polFinalSamplesRefined[i] = XF[ni];
        samplesPolynomialRefined.push_back(XF);
    }
    dlmwrite("../../results/planarLowThrust/outputPropagatedSamplesRefinedPoly.txt", X0F);
    X0F.clear();


    // Compute RMSE errors
    vector<vector<double>> RMSEFable(ni + 1, vector<double>(ns, 0.0));
    vector<vector<double>> RMSERefined(ni + 1, vector<double>(ns, 0.0));
    for (unsigned int j = 0; j <= ni; j++)
    {
        for (unsigned int k = 0; k < ns; k++)
        {
            // Compute errors square
            for (unsigned int i = 0; i <= nSamples; i++)
            {
                RMSEFable[j][k] += pow(samplesDoubleFable[i][j][k] - samplesPolynomialFable[i][j][k], 2.0);
                RMSERefined[j][k] += pow(samplesDoubleRefined[i][j][k] - samplesPolynomialRefined[i][j][k], 2.0);
            }

            // Normalise and take square root
            RMSEFable[j][k] /= (double) (nSamples + 1);
            RMSEFable[j][k] = sqrt(RMSEFable[j][k]);
            RMSERefined[j][k] /= (double) (nSamples + 1);
            RMSERefined[j][k] = sqrt(RMSERefined[j][k]);
        }
    }

    // Print samples evolution for fable control
    for (unsigned int i = 0; i <= nSamples; i++)
    {
        if (i == 0)
        {
            cout << endl << "Fable double vs Poly vs Final error = " << endl;
            for (unsigned int j = 0; j <= ni; j++) 
            {
                
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesDoubleFable[i][j][k] << " ";
                cout << "; y = " << doubleYFable[i];
                cout << endl;
                
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesPolynomialFable[i][j][k] << " ";
                cout << "; y = " << polYFable[i];
                
                cout << endl;
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesDoubleFable[i][j][k] - samplesPolynomialFable[i][j][k] << " ";
                cout << endl << endl;
                
            }
            cout << endl;
        }
    }

    
    // Print samples evolution for refined control
    for (unsigned int i = 0; i <= nSamples; i++)
    {
        if (i == 0)
        {
            cout << endl << "Refined double vs Poly vs Final error = " << endl;
            for (unsigned int j = 0; j <= ni; j++) 
            {
                
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesDoubleRefined[i][j][k] << " ";
                cout << "; y = " << doubleYRefined[i];
                cout << endl;
                
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesPolynomialRefined[i][j][k] << " ";
                cout << "; y = " << polYRefined[i];
                
                cout << endl;
                for (unsigned int k = 0; k < ns; k++)
                    cout << samplesDoubleRefined[i][j][k] - samplesPolynomialRefined[i][j][k] << " ";
                cout << endl << endl;
                
            }
            cout << endl;
        }
    }

    
    
    cout << endl << "RMSEFable of each state (column) for different times (row) = " << endl;
    for ( unsigned int j = 0 ; j <=ni; j++ )
    {
        cout << "L = " << L_guess[j] << " RMSEFable = " ;
        cout << RMSEFable[j] << endl;
    }
    cout << endl;
    cout << "RMSERefined of each state (column) for different times (row) = " << endl;
    for ( unsigned int j = 0 ; j <=ni; j++ )
    {
        cout << "L = " << L_guess[j] << " RMSERefined = " ;
        cout << RMSERefined[j] << endl;
    }
    cout << endl;

    
    cout << "Fable double Final Probability = " ;
    constraintParams.pSamples = &doubleFinalSamplesFable;
    cout << finalConstraint.evaluate() << endl;

    cout << "Fable polynomial Final Probability = " ;
    constraintParams.pSamples = &polFinalSamplesFable;
    finalConstraint.evaluate();
    cout << finalConstraint.evaluate() << endl;

    cout << "Refined double Final Probability = " ;
    constraintParams.pSamples = &doubleFinalSamplesRefined;
    cout << finalConstraint.evaluate() << endl;

    cout << "Refined polynomial Final Probability = " ;
    constraintParams.pSamples = &polFinalSamplesRefined;
    finalConstraint.evaluate();
    cout << finalConstraint.evaluate() << endl;


    cout << "Final Sigma Region: " << endl;
    cout << "LB = ";
    for (unsigned int i = 0 ; i < ns; i++ ){
        cout << finalLB[i] << " ";
    }
    cout << endl;
    cout << "UB = ";
    for (unsigned int i = 0 ; i < ns; i++ ){
        cout << finalUB[i] << " ";
    }
    cout << endl;
    cout << endl;



    return 0;
}




int main(){

    cout << endl << "Welcome to the IPANEMA Planar Low Thrust example!" << endl;

    equinoctialPlanar();

//    reintegrateEquinoctialPlanar();

    cout << endl << "Goodbye from the IPANEMA Planar Low Thrust example!" << endl;

    return 0;
}