/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include "intrusivePolynomialPropagator.h"

using namespace ipanema;
using namespace propagatorSurrogates;


/**
 * Constructor
 */
intrusivePolynomialPropagator::intrusivePolynomialPropagator(const intrusivePolynomialPropagatorParams* pParams) :
        m_pParams(pParams)
{
    // Nothing here
}


// Build the surrogate
void intrusivePolynomialPropagator::build()
{
    // If already saved
    if (m_pParams->readFromFile)
    {
        buildByFile();
    }
    else
    {
        if (m_pParams->usePredefinedRange && m_pParams->isSystemAutomonous)
            buildByNewAutonomousPropagationWithCustomRange();
        else
            buildByNewPropagation();
    }

} // build


// Evaluate the surrogate
vector<double> intrusivePolynomialPropagator::propagate(const vector<double>& x) const
{
    // Problem dimension
    unsigned int ns = (*m_pParams->pStartPolynomials).size();                          // State size
    unsigned int nT = (*m_pParams->pTimeGrid).size();                                  // Time grid size
    unsigned int nc = (*m_pParams->pControlCoefficients)[0].size();                    // Control size
    unsigned int nVar = (unsigned int) (*m_pParams->pStartPolynomials)[0].get_nvar();  // Number of polynomial variables

    // Sanity check on initial condition
    if (x.size()!=ns)
        ipanema_throw("Input vector of wrong size");

    // Modifiable initial and final condition
    vector<double> x0(x), xf(ns);
    // Buffer scaled state
    vector<double> scaledX(nVar,0.0);

    // Polynomial propagation
    for (unsigned int i = 0 ; i < nT-1; i++)
    {
        // Set and scale state
        for ( unsigned int j = 0 ; j < ns; j++ )
        {
            // Scale
            scaledX[j] = (x0[j]-m_rangePolynomials[i][j][0]) /
                                 (m_rangePolynomials[i][j][1]-m_rangePolynomials[i][j][0]) * 2.0 - 1.0;

            // Sanity check
            if(abs(scaledX[j])>1.0)
            {

                ipanema_throw("index = " + to_string(j) + "; nonScaledX = " + to_string(x0[j]) + " scaledX = " + to_string(scaledX[j]) +
                              "; bounds = [" + to_string(m_rangePolynomials[i][j][0]) + "," +
                              to_string(m_rangePolynomials[i][j][1])+ "]");
            }
        }

        // Set to scaled control
        for ( unsigned int j = 0 ; j < nc; j++ )
        {
            // Scale
            scaledX[j+ns] = ((*m_pParams->pControlCoefficients)[i][j]-m_pParams->controlBounds[j][0]) /
                            (m_pParams->controlBounds[j][1]-m_pParams->controlBounds[j][0]) * 2.0 - 1.0;

            // Sanity check
            if(abs(scaledX[j+ns])>1.0)
            {
                ipanema_throw("index = " + to_string(j) + "; nonScaledU = "
                              + to_string((*m_pParams->pControlCoefficients)[i][j]) +
                                      " scaledU  = " + to_string(scaledX[j+ns]) +
                              "; bounds = [" + to_string(m_pParams->controlBounds[j][0]) + "," +
                              to_string(m_pParams->controlBounds[j][1]) + "]");
            }
        }

        // Evaluate point for each state surrogate
        for ( unsigned int j = 0 ; j < ns; j++ )
        {
            xf[j] = m_propagatedPolynomials[i+1][j].evaluate(scaledX);
        }

        // Set initial point to final response
        x0 = xf;
    }

    return xf;
}


// Evaluate the surrogate
vector<vector<double>> intrusivePolynomialPropagator::propagateMiddle (const vector<double>& x) const
{
    // Problem dimension
    unsigned int ns = (*m_pParams->pStartPolynomials).size();                          // State size
    unsigned int nT = (*m_pParams->pTimeGrid).size();                                  // Time grid size
    unsigned int nc = (*m_pParams->pControlCoefficients)[0].size();                    // Control size
    unsigned int nVar = (unsigned int) (*m_pParams->pStartPolynomials)[0].get_nvar();  // Number of polynomial variables

    vector<vector<double>> XF;

    // Sanity check on initial condition
    if (x.size()!=ns)
        ipanema_throw("Input vector of wrong size");

    // Modifiable initial and final condition
    vector<double> x0(x), xf(ns);
    // Buffer scaled state
    vector<double> scaledX(nVar,0.0);

    // Polynomial propagation
    for (unsigned int i = 0 ; i < nT; i++) // Starts from 1 as i==0 is polynomial at time 0
    {
        // Set and scale state
        for ( unsigned int j = 0 ; j < ns; j++ )
        {
            if (i>0)
            {
                // Scale
                scaledX[j] = (x0[j]-m_rangePolynomials[i-1][j][0]) /
                             (m_rangePolynomials[i-1][j][1]-m_rangePolynomials[i-1][j][0]) * 2.0 - 1.0;
            } else {
                // Scale
                scaledX[j] = (x0[j]-m_rangePolynomials[0][j][0]) /
                             (m_rangePolynomials[0][j][1]-m_rangePolynomials[0][j][0]) * 2.0 - 1.0;
            }

            // Sanity check
            if(abs(scaledX[j])>1.0)
            {
                ipanema_throw("index = " + to_string(j) + "; nonScaledX = " + to_string(x0[j]) + " scaledX = " + to_string(scaledX[j]) +
                              "; bounds = [" + to_string(m_rangePolynomials[i][j][0]) + "," +
                              to_string(m_rangePolynomials[i][j][1])+ "]");
            }
        }

        // Set to scaled control
        if (i>0)
        {
            for ( unsigned int j = 0 ; j < nc; j++ )
            {
                // Scale
                scaledX[j+ns] = ((*m_pParams->pControlCoefficients)[i-1][j]-m_pParams->controlBounds[j][0]) /
                                (m_pParams->controlBounds[j][1]-m_pParams->controlBounds[j][0]) * 2.0 - 1.0;

                // Sanity check
                if(abs(scaledX[j+ns])>1.0)
                {
                    ipanema_throw("index = " + to_string(j) + "; nonScaledU = "
                                  + to_string((*m_pParams->pControlCoefficients)[i-1][j]) +
                                  " scaledU  = " + to_string(scaledX[j+ns]) +
                                  "; bounds = [" + to_string(m_pParams->controlBounds[j][0]) + "," +
                                  to_string(m_pParams->controlBounds[j][1]) + "]");
                }
            }
        }

        // Evaluate point for each state surrogate
        for ( unsigned int j = 0 ; j < ns; j++ )
        {
            xf[j] = m_propagatedPolynomials[i][j].evaluate(scaledX);
        }

        XF.push_back(xf);

        // Set initial point to final response
        x0 = xf;
    }

    return XF;
}


/**
 * Help routines to shorten code
 */

void intrusivePolynomialPropagator::computePolynomialRange ( const vector<smartuq::polynomial::chebyshev_polynomial>* pPolynomials,
                                                             vector<vector<double>> &rRange )
{

    // Number of Polynomials
    unsigned int nPol = pPolynomials->size();

    // Clear and reserve space for output range
    rRange.clear();
    rRange.resize(nPol);

    for (unsigned int i = 0 ; i < nPol; i++)
    {
        if ( m_pParams->usePredefinedRange )
        {
            if ( m_pParams->predefinedRange[i].size()!=2 )
                ipanema_throw("Error in predefined range size");

            rRange[i] = m_pParams->predefinedRange[i];
        }
        else if ( m_pParams->useConservativeRange )
        {
            rRange[i] = m_propagatedPolynomials[0][i].get_range();
        }
        else
        {
            rRange[i] = m_propagatedPolynomials[0][i].get_range(m_pParams->nSamplesMCPolynomialRange);

            // Add percentage correction
            double halfRange = (rRange[i][1]-rRange[i][0])/2.0;
            rRange[i][0] -= abs(m_pParams->percentageSafeMCPolynomialRange*halfRange);
            rRange[i][1] += abs(m_pParams->percentageSafeMCPolynomialRange*halfRange);
        }

        // Sanity checks
        if ( rRange[i].size()!=2 )
            ipanema_throw("Range size is not 2");
        if ( (rRange[i][1]-rRange[i][0]) < 0.0 )
            ipanema_throw("Upper range is smaller than lower range");
    }

} // computePolynomialRange



/**
 * Reinitialise polynomials as hyperbox given range
 */

void intrusivePolynomialPropagator::reinitialisePolynomialByRange (const vector<vector<double>>* pRange,
                                                                   vector<smartuq::polynomial::chebyshev_polynomial>* pPolynomials) const
{
    // Number of polynomials
    unsigned int nPol = pPolynomials->size();

    // Sanity check on dimension
    if (pRange->size()!=nPol)
        ipanema_throw("Number of polynomials and range vector do not coincide");

    // Buffer for coefficients
    unsigned int nCoeffs = (*pPolynomials)[0].get_coeffs().size();
    vector<double> coeffs(nCoeffs,0.0);

    // Assign new coefficients to create hyper-box
    for ( unsigned int i = 0 ; i < nPol; i++ )
    {
        // Set new coefficients
        coeffs = vector<double>(nCoeffs,0.0);
        coeffs[0]   = ((*pRange)[i][1]+(*pRange)[i][0])/2.0;
        coeffs[i+1] = ((*pRange)[i][1]-(*pRange)[i][0])/2.0;

        // Assign new coefficients to polynomial
        (*pPolynomials)[i].set_coeffs(coeffs);
    }

} // reinitialisePolynomialByRange



/**
 * Build surrogate by intrusive propagation
 */
void intrusivePolynomialPropagator::buildByNewPropagation()
{
    // Clear output vectors
    m_propagatedPolynomials.clear();
    m_rangePolynomials.clear();

    // State size
    unsigned int ns = m_pParams->pStartPolynomials->size();

    // Time grid size
    unsigned int nT = (*m_pParams->pTimeGrid).size();

    // Number of integration steps per unit
    double nStepsPerUnit = ((double)m_pParams->nIntegrationSteps)/
                           ((*m_pParams->pTimeGrid)[nT-1]-(*m_pParams->pTimeGrid)[0]);

    // Buffer for integration
    vector<smartuq::polynomial::chebyshev_polynomial> X0(*(m_pParams->pStartPolynomials));
    vector<smartuq::polynomial::chebyshev_polynomial> XF;
    double t0, tf;

    // Reserve space for output value
    m_propagatedPolynomials.resize(nT);
    m_rangePolynomials.resize(nT);

    // Polynomial for initial time is X0
    m_propagatedPolynomials[0] = X0;
    m_rangePolynomials[0].resize(ns);
    // Assign range of first polynomial directly by first assignment
    for (unsigned int i = 0 ; i < ns ; i++)
        m_rangePolynomials[0][i] = m_propagatedPolynomials[0][i].get_range();

    // Propagate through dynamics
    cout << endl << endl << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% POLYNOMIAL PROPAGATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl << endl;
    for ( unsigned int i = 1 ; i < nT ; i++ )
    {
        // Print iteration number
        cout << "Iteration " << to_string(i) << "/" << to_string(nT-1) << endl;

        // Initial and final times
        t0 = (*m_pParams->pTimeGrid)[i-1];
        tf = (*m_pParams->pTimeGrid)[i];

        // Number of integration steps for current interval
        auto nStep_i = (int)ceil(nStepsPerUnit*(tf-t0));

        // Integrate
        m_pParams->pIntegrator->integrate(t0,tf,nStep_i,X0,XF);

        // Save XF in output vector
        m_propagatedPolynomials[i] = XF;

        // Compute and save range in output vector
        computePolynomialRange(&m_propagatedPolynomials[i],m_rangePolynomials[i]);

        // Set new initial state as hyper-cube
        reinitialisePolynomialByRange(&m_rangePolynomials[i],&X0);

    } // for cycle over time

    // If requested, print range and polynomial coefficients to file
    if ( m_pParams->printToFile )
        printBuildToFile();

}



/**
 * Build surrogate by intrusive propagation
 */
void intrusivePolynomialPropagator::buildByNewAutonomousPropagationWithCustomRange()
{
    // Check on autonomous system
    if (!m_pParams->isSystemAutomonous)
        ipanema_throw("Condition of autonomous system not met");
    
    // Clear output vectors
    m_propagatedPolynomials.clear();
    m_rangePolynomials.clear();

    // Dimension State size
    auto ns = m_pParams->pStartPolynomials->size(), nT = (*m_pParams->pTimeGrid).size();

    // Number of integration steps per unit
    auto nStepsPerUnit = ((double)m_pParams->nIntegrationSteps)/
                         ((*m_pParams->pTimeGrid)[nT-1]-(*m_pParams->pTimeGrid)[0]);

    // Buffer for integration
    vector<smartuq::polynomial::chebyshev_polynomial> X0(*(m_pParams->pStartPolynomials)), XF;
    double t0, tf;

    // Reserve space for output value
    m_propagatedPolynomials.resize(nT);
    m_rangePolynomials.resize(nT);

    // Polynomial for initial time is X0
    m_propagatedPolynomials[0] = X0;
    m_rangePolynomials[0].resize(ns);
    // Assign range of first polynomial directly by first assignment
    for (unsigned int i = 0 ; i < ns ; i++)
        m_rangePolynomials[0][i] = m_propagatedPolynomials[0][i].get_range();

    // Propagate first polynomial
    cout << endl << endl << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% POLYNOMIAL PROPAGATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl << endl;
    {
        // Print iteration number
        cout << "Iteration " << to_string(1) << "/" << to_string(nT-1) ;//<< endl;

        // Initial and final times
        t0 = (*m_pParams->pTimeGrid)[0];
        tf = (*m_pParams->pTimeGrid)[1];

        // Number of integration steps for current interval
        auto nStep_i = (int)ceil(nStepsPerUnit*(tf-t0));

        // Integrate
        m_pParams->pIntegrator->integrate(t0,tf,nStep_i,X0,XF);

        // Save XF in output vector
        m_propagatedPolynomials[1] = XF;

        // Compute and save range [it will use custom] in output vector
        computePolynomialRange(&m_propagatedPolynomials[1],m_rangePolynomials[1]);

        // Set new initial state as hyper-cube with custom range
        reinitialisePolynomialByRange(&m_rangePolynomials[1],&X0);
    }
    
    // Propagate next steps starting from custom range 
    // 
    // Compute ordered relative times wrt second time step
    vector<pair<unsigned int, double>> relTimesOrderedWithIndexExceptFirst(nT-2);
    for (auto i = 0 ; i < nT-2; i++) {
        relTimesOrderedWithIndexExceptFirst[i].first  = i+2;
        relTimesOrderedWithIndexExceptFirst[i].second = (*m_pParams->pTimeGrid)[i+2] - (*m_pParams->pTimeGrid)[i+1];
    }

    // Sort by time
    //
    // Compare function
    function<bool(pair<unsigned int, double> & , pair<unsigned int, double> & )> comparePair =
            []( pair<unsigned int, double> &a, pair<unsigned int, double> &b) -> bool {
                return (a.second < b.second);
            };

    // Sort algorithm
    sort(relTimesOrderedWithIndexExceptFirst.begin(),
         relTimesOrderedWithIndexExceptFirst.end(),
         comparePair);
    
    // Propagate through dynamics
    //
    // Set initial and final time [autonomous system]
    t0 = (*m_pParams->pTimeGrid)[1];
    tf = t0+relTimesOrderedWithIndexExceptFirst[0].second;
    
    // Iterate
    for ( unsigned int i = 0 ; i < nT-2 ; i++ )
    {
        // Print iteration number
        cout << "Iteration " << to_string(i+2) << "/" << to_string(nT-1) << endl;

        // Number of integration steps for current interval
        auto nStep_i = (int)ceil(nStepsPerUnit*(tf-t0));
        auto index = relTimesOrderedWithIndexExceptFirst[i].first;

        // Integrate
        m_pParams->pIntegrator->integrate(t0,tf,nStep_i,X0,XF);

        // Save XF in output vector
        m_propagatedPolynomials[index] = XF;

        // Compute and save range in output vector
        computePolynomialRange(&m_propagatedPolynomials[index],m_rangePolynomials[index]);

        // Exclude last iteration
        if (i!=nT-3) {

            // Set new initial state to final state without reinitialising as they all start from same hypercube and time has been ordered
            X0 = XF;

            // New initial and final time
            t0 = tf;
            tf += (relTimesOrderedWithIndexExceptFirst[i+1].second-relTimesOrderedWithIndexExceptFirst[i].second);
        }
        
    } // for cycle over time

    // If requested, print range and polynomial coefficients to file
    if ( m_pParams->printToFile )
        printBuildToFile();
}

/**
 * Build surrogate by input file
 */

void intrusivePolynomialPropagator::buildByFile()
{
    // Clear output vectors
    m_propagatedPolynomials.clear();
    m_rangePolynomials.clear();

    // State size
    unsigned int ns = m_pParams->pStartPolynomials->size();

    // Time grid size
    unsigned int nT = (*m_pParams->pTimeGrid).size();

    // Polynomial quantities
    unsigned int nDeg = (unsigned int) ((*m_pParams->pStartPolynomials)[0].get_degree());
    unsigned int nCoeffs = (*m_pParams->pStartPolynomials)[0].get_coeffs().size();

    /**
     * Input range
     */
    // Reserve space for output value
    m_rangePolynomials = vector<vector<vector<double>>> (nT,
                                                         vector<vector<double>>(ns,
                                                                                vector<double>(2)
                                                         )
    );

    // Read polynomial ranges from file
    vector<vector<double>> rangeInput;
    dlmread(m_pParams->inputFilePrefix+"polRangeDeg"+to_string(nDeg)+".txt", rangeInput);

    // Sanity check
    if (rangeInput.size()!=nT)
        ipanema_throw("Input range read from file has not correct time size");

    // Assign input to class member
    for (unsigned int i = 0; i < nT; i++)
    {
        // Sanity check on dimension
        if (rangeInput[i].size()!=1+2*ns)
            ipanema_throw("Input range row has not correct size: it should be (independent variable)+2*rangeOfJVariable");
        // Sanity check on time
        if (abs(rangeInput[i][0]-(*m_pParams->pTimeGrid)[i])>1.0e-6)
            ipanema_throw("Input range row have not matching independent variable");

        // Assign range to m_rangePolynomials
        for (unsigned int j = 0; j < ns; j++)
        {
            // Lower bound
            m_rangePolynomials[i][j][0] = rangeInput[i][1+j*2];
            // Upper bound
            m_rangePolynomials[i][j][1] = rangeInput[i][1+j*2+1];
        }

    }
    rangeInput.clear();


    /**
     * Input polynomial coefficients
     */

    // Reserve space for output value by copying nT times initial startPolynomials
    m_propagatedPolynomials =
            vector<vector<smartuq::polynomial::chebyshev_polynomial>>(nT,
                        vector<smartuq::polynomial::chebyshev_polynomial>(*m_pParams->pStartPolynomials));

    // Read polynomial coefficients from file
    vector<vector<double>> polInput;
    dlmread(m_pParams->inputFilePrefix+"polCoeffDeg"+to_string(nDeg)+".txt", polInput);

    // Sanity check
    if (polInput.size()!=nT)
        ipanema_throw("Input range read from file has not correct time size");

    // Assign input to class member
    for (unsigned int i = 0; i < polInput.size(); i++)
    {

        // Sanity check on dimension
        if (polInput[i].size()!=1+nCoeffs*ns)
            ipanema_throw("Input range row has not correct size: it should be (independent variable)+2*rangeOfJVariable");
        // Sanity check on time
        if (abs(polInput[i][0]-(*m_pParams->pTimeGrid)[i])>1.0e-6)
            ipanema_throw("Input range row have not matching independent variable");

        // Assign coefficients to polynomials
        vector<double> coeff_ij(nCoeffs);
        for (unsigned int j = 0; j < ns; j++)
        {
            // Input coefficients
            for (unsigned int k = 0 ; k < nCoeffs; k++)
            {
                coeff_ij[k] = polInput[i][1+j*nCoeffs+k];
            }

            // Assign coefficients if not first polynomial
//            if ( i!=0 )
//            {
                m_propagatedPolynomials[i][j].set_coeffs(coeff_ij);
//            }
//            else // Sanity check on initial polynomials to coincide with input files
//            {
//                vector<double> checkCoeff = m_propagatedPolynomials[i][j].get_coeffs();
//
//                for (unsigned int k = 0 ; k < nCoeffs; k++)
//                {
//                    if (abs(checkCoeff[k]-coeff_ij[k])>1.0e-6)
//                    {
//                        ipanema_throw("Input starting polynomial coefficients and example defined ones do not match");
//                    }
//                }
//            }
        }
    }
    polInput.clear();

}



/**
 * Print polynomial coefficients and range to file
 */

void intrusivePolynomialPropagator::printBuildToFile() const
{

    // State size
    unsigned int ns = m_pParams->pStartPolynomials->size();

    // Time grid size
    unsigned int nT = (*m_pParams->pTimeGrid).size();

    // Polynomial quantities
    unsigned int nDeg = (unsigned int) ((*m_pParams->pStartPolynomials)[0].get_degree());
    unsigned int nCoeffs = (*m_pParams->pStartPolynomials)[0].get_coeffs().size();

    // Print range to file
    vector<vector<double>> rangeOut (nT,
                                               vector<double>(1+2*ns));
    for (unsigned int i = 0 ; i < nT; i++)
    {
        // Independent variable grid
        rangeOut[i][0] = (*m_pParams->pTimeGrid)[i];

        // Polynomial ranges
        for (unsigned int j = 0 ; j < ns; j++)
        {
            // Lower bound
            rangeOut[i][1+j*2]   = m_rangePolynomials[i][j][0];
            // Upper bound
            rangeOut[i][1+j*2+1] = m_rangePolynomials[i][j][1];
        }
    }
    // Print ranges to file
    dlmwrite(m_pParams->outputFilePrefix+"polRangeDeg"+to_string(nDeg)+".txt",rangeOut);
    rangeOut.clear();


    // Polynomial coefficients struct to be printed
    vector<vector<double>> polOut (nT,
                                             vector<double>(1+nCoeffs*ns));
    for (unsigned int i = 0 ; i < nT; i++)
    {
        // Independent variable grid
        polOut[i][0] = (*m_pParams->pTimeGrid)[i];

        // Polynomial ranges
        vector<double> coeff_ij(nCoeffs);
        for (unsigned int j = 0 ; j < ns; j++)
        {
            // Coeffs from polynomials
            coeff_ij = m_propagatedPolynomials[i][j].get_coeffs();

            // Set to output struct
            for ( unsigned int k = 0 ; k < nCoeffs; k++ )
            {
                polOut[i][1+j*nCoeffs+k] = coeff_ij[k];
            }
        }
    }
    // Print polynomial coefficients to file
    dlmwrite(m_pParams->outputFilePrefix+"polCoeffDeg"+to_string(nDeg)+".txt",polOut);
    polOut.clear();

}


// Sanity checks on input parameters
void intrusivePolynomialPropagator::inputParamsSanityCheck() const
{
    // Input polynomials cannot be empty
    if((*m_pParams->pStartPolynomials).empty())
        ipanema_throw("Initial polynomials vector empty");

    // Problem dimension
    unsigned int ns = m_pParams->pStartPolynomials->size();                            // State size
    unsigned int nc = (*m_pParams->pControlCoefficients)[0].size();                    // Control size
    unsigned int nVar = (unsigned int) (*m_pParams->pStartPolynomials)[0].get_nvar();  // Number of polynomial variables

    // Sanity check (no dynamical uncertain parameters allowed for now)
    if(nVar!=(ns+nc))
        ipanema_throw("No dynamical uncertain parameters allowed for now");
}