/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#ifndef IPANEMA_INTRUSIVEPOLYNOMIALPROPAGATOR_H
#define IPANEMA_INTRUSIVEPOLYNOMIALPROPAGATOR_H

#include <utility>
#include <algorithm>

#include "propagatorSurrogate.h"

#include "ipanemaUtils/utils.h"
//#include "Polynomial/chebyshev_polynomial.h"
#include "Polynomial/chebyshev.h"
#include "Integrators/base_integrator.h"


namespace ipanema
{
namespace propagatorSurrogates
{

    class intrusivePolynomialPropagator : public propagatorSurrogate {

        /**
         * Parameters
         */

    public:

        struct intrusivePolynomialPropagatorParams : public propagatorSurrogateParams {

            // Initial polynomials to propagate
            const std::vector<
                    smartuq::polynomial::chebyshev_polynomial>*     pStartPolynomials;

            // Integrator (with dynamics already included)
            const smartmath::integrator::base_integrator<
                    smartuq::polynomial::chebyshev_polynomial>*     pIntegrator;

            // Time vector at which propagation points shall be saved
            std::vector<double>*                                    pTimeGrid;

            // Control parematers
            std::vector<std::vector<double>>*                       pControlCoefficients;

            // Control bounds
            std::vector<std::vector<double>>                        controlBounds;

            // Number of integration steps for whole time grid
            unsigned int                                            nIntegrationSteps;

            // Flag for predefined range [highest priority]
            bool                                                    usePredefinedRange;

            // Predefined Range
            std::vector<std::vector<double>>                        predefinedRange;

            // Flag for getting polynomial range by conservative formula [if usePredefinedRange = false]
            bool                                                    useConservativeRange;

            // Number of samples for polynomial range by MC [if conservativePolynomialRange = false]
            unsigned int                                            nSamplesMCPolynomialRange;

            // Safe percentage for range estimation by MC (shift by halfRange*percentageSafeMCPolynomialRange) [if conservativePolynomialRange = false]
            double                                                  percentageSafeMCPolynomialRange;

            // Flag for printing output file propagated polynomials and range
            bool                                                    printToFile;

            // Autonomous system except control
            bool                                                    isSystemAutomonous = false;

            // Output file prefix for propagated polynomials and range [if printToFile = true]
            std::string                                             outputFilePrefix;

            // Flag for getting from file propagated polynomials and range
            bool                                                    readFromFile;

            // Input file prefix for propagated polynomials and range [if readFromFile = true]
            std::string                                             inputFilePrefix;

        };

        /**
         * Member list
         */

    protected:

        // Parameters
        const intrusivePolynomialPropagatorParams*              m_pParams;

        // Propagated polynomial surrogates at each time in pTimeGrid
        std::vector<std::vector<
                smartuq::polynomial::chebyshev_polynomial>>     m_propagatedPolynomials;

        // Range of polynomials at each time in pTimeGrid used to scale during propagation
        std::vector<std::vector<std::vector<double>>>           m_rangePolynomials;


        /**
         * Constructor
         */

    public:

        // Default constructor
        intrusivePolynomialPropagator(const intrusivePolynomialPropagatorParams* pParams);



        /**
         * List of virtual functions to be implemented in derived classes
         */

    public:

        // Evaluate the surrogate
        std::vector<double> propagate (const std::vector<double>& x) const;

        // Evaluate the surrogate
        std::vector<std::vector<double>> propagateMiddle (const std::vector<double>& x) const;

        // Build the surrogate
        void build();


        /**
         * Help routines to shorten code
         */

    private:

        // Compute polynomial range
        void computePolynomialRange (const std::vector<smartuq::polynomial::chebyshev_polynomial>* pPolynomials,
                                     std::vector<std::vector<double>> &rRange ) ;

        // Reinitialise polynomials as hyperbox given range
        void reinitialisePolynomialByRange (const std::vector<std::vector<double>>* pRange,
                                            std::vector<smartuq::polynomial::chebyshev_polynomial>* pPolynomials) const;

        // Build surrogate by intrusive propagation
        void buildByNewPropagation();

        // Build surrogate by intrusive propagation with custom range for autonomous systems only
        void buildByNewAutonomousPropagationWithCustomRange();

        // Build surrogate by input file
        void buildByFile();

        // Print polynomial coefficients and range to file
        void printBuildToFile() const;


    protected:

        // Sanity checks on input parameters
        void inputParamsSanityCheck() const;

    };

} // namespace propagatorSurrogates
} // namespace ipanema

#endif //IPANEMA_INTRUSIVEPOLYNOMIALPROPAGATOR_H
