/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include "samplingProbabilityConstraint.h"

using namespace std;
using namespace ipanema;
using namespace ipanemaConstraints;


// Default constructor
samplingProbabilityConstraint::samplingProbabilityConstraint(const samplingConstraintsParams* pParams)
        : samplingConstraints(pParams),
          m_pParams(static_cast<const samplingProbabilityConstraintParams*>(pParams))
{
    // Sanity check
    if ( m_pParams == nullptr )
        ipanema_throw("Dynamic cast failed: " +
                              "input parameter struct was not pointing to samplingProbabilityConstraintParams");
    if (!m_pParams->expectationMC)
    {
        if ( m_pParams->pSamplesDensities == nullptr )
            ipanema_throw("m_pParams->pSamplesDensities null pointer");

        if ( (*m_pParams->pSamples).size() != (*m_pParams->pSamplesDensities).size() )
            ipanema_throw("Number of samples differs from number of sample densities");
    }
}




// Evaluate the constraint in pSamples
double samplingProbabilityConstraint::evaluate()
{
    // Probability of final region to belong to set
    double valueSetProbability = computeSamplesSetProbability();

    // Sanity check
    if ( valueSetProbability<0.0 || valueSetProbability>1.0 )
        ipanema_throw("Computed value is not probability measure");

    // As constraints usually imposed as g<=0.0, we return probability of not belonging to set
    return (1.0-valueSetProbability);
}


// Compute probability of pSamples to be in set defined by pSetCondition
double samplingProbabilityConstraint::computeSamplesSetProbability()
{
    // Output value
    double samplesSetProbability = 0.0;

    // Number of samples
    unsigned int nTotSamples = (*m_pParams->pSamples).size();

    // Set Membership value for each sample
    std::vector<double> valueSamplesSetMembership (nTotSamples,0.0);

    //// Evaluate set membership value
    //
    // Evaluate full set
    if (!m_pParams->sampleSubsetOption || (*m_pParams->pUseFullSampleSet))
    {
        m_subsetIndex.clear();
        for (unsigned int i = 0; i < nTotSamples; i++)
        {
            // Compute value of sample
            valueSamplesSetMembership[i] =
                    computeSampleSetMembership((*m_pParams->pSamples)[i]);

            // Save subset of samples with positive membership if requested
            if(m_pParams->sampleSubsetOption)
            {
                if (valueSamplesSetMembership[i] > 0.0)
                {
                    m_subsetIndex.push_back(i);
                }
            }
        }
    }
    // Evaluate subset
    else
    {
        for (const unsigned int i : m_subsetIndex)
        {
            // Compute value of sample
            valueSamplesSetMembership[i] =
                    computeSampleSetMembership((*m_pParams->pSamples)[i]);
        }
    }

    /**
     * Compute probability as Expectation of membership function
     */

    // MonteCarlo approximation of Set Mean
    if (m_pParams->expectationMC)
    {
        for (unsigned int i = 0 ; i < nTotSamples; i++)
        {
            samplesSetProbability += valueSamplesSetMembership[i];
        }
        samplesSetProbability /= (double) nTotSamples;
    }
    // Quadrature approximation of expectation integral (using probability densities as weights)
    else
    {
        for (unsigned int i = 0 ; i < nTotSamples; i++)
        {
            samplesSetProbability +=
                    (*m_pParams->pSamplesDensities)[i] * valueSamplesSetMembership[i];
        }
    }

//    std::cout << std::endl << "samplesSetProbability = " << samplesSetProbability << std::endl;
//    std::cout << std::endl;
//    ipanema_throw("stop here");

    // Return output value
    return samplesSetProbability;
}


// Compute membership of pSamples to be in set defined by pSetCondition
double samplingProbabilityConstraint::computeSampleSetMembership(const std::vector<double> &sample)
{
    // Number of samples
    double nSamples = (double)(*m_pParams->pSamples).size();

    // Output value
    double valueMembershipSample = 0.0;

    // Evaluate set value membership of sample
    double valueSetSample = m_pParams->pSetCondition(sample);

    // Scale valueSetSample in [-1,+1]
    valueSetSample = (valueSetSample-m_pParams->setValueLB) /
                     (m_pParams->setValueUB-m_pParams->setValueLB) * 2.0 - 1.0;


    double r = 1.3*std::pow(nSamples,-1.0/5.0);
    switch (m_pParams->membershipType)
    {
        case INDICATOR_MEMBERSHIP:
            if ( std::abs(valueSetSample)<=1.0 )
                valueMembershipSample = 1.0;
            break;
        case LINEAR_MEMBERSHIP:
            m_fIntegrate = [xx=valueSetSample,rr=r](double y) -> double {

                double pp = (xx-y)/rr;

                double res = 0.0;
                if ( std::abs(pp)<=1.0 )
                {
                    res = 1.0 - std::abs(pp);
                    res /= rr;
                }

                return res; };
//                valueMembershipSample = 1.0-std::abs(valueSetSample);
            break;
        case TRIGONOMETRIC_MEMBERSHIP:
            m_fIntegrate = [xx=valueSetSample,rr=r](double y) -> double {

                double pp = (xx-y)/rr;

                double res = 0.0;
                if ( std::abs(pp)<=1.0 )
                {
                    res = M_PI*std::cos(M_PI*pp/2.0)/4.0;
                    res /= rr;
                }

                return res; };
//                valueMembershipSample = M_PI*std::cos(M_PI*valueSetSample/2.0)/4.0;
            break;
        case QUADRATIC_MEMBERSHIP:
            m_fIntegrate = [xx=valueSetSample,rr=r](double y) -> double {
                double pp = (xx-y)/rr;

                double res = 0.0;
                if ( std::abs(pp)<=1.0 )
                {
                    res = 3.0*(1.0-std::pow(pp,2.0))/4.0;
                    res /= rr;
                }

                return res; };
//                valueMembershipSample = 3.0*(1.0-std::pow(valueSetSample,2.0))/4.0;
            break;
        case BIQUADRATIC_MEMBERSHIP:
            m_fIntegrate = [xx=valueSetSample,rr=r](double y) -> double {
                double pp = (xx-y)/rr;

                double res = 0.0;
                if ( std::abs(pp)<=1.0 )
                {
                    res = 15.0 * std::pow((1.0 - std::pow(pp, 2.0)), 2.0) / 16.0;
                    res /= rr;
                }

                return res; };
//                valueMembershipSample = 15.0*std::pow((1.0-std::pow(valueSetSample,2.0)),2.0)/16.0;
            break;
        case TRIQUADRATIC_MEMBERSHIP:
            m_fIntegrate = [xx=valueSetSample,rr=r](double y) -> double {
                double pp = (xx-y)/rr;

                double res = 0.0;
                if ( std::abs(pp)<=1.0 )
                {
                    res = 35.0 * std::pow((1.0 - std::pow(pp, 2.0)), 3.0) / 32.0;
                    res /= rr;
                }

                return res; };
//                valueMembershipSample = 35.0*std::pow((1.0-std::pow(valueSetSample,2.0)),3.0)/32.0;
            break;
        default:
            ipanema_throw("Membership type ill-defined");
            break;
    }

    // Integrate according to mollifier
    if ( m_pParams->membershipType != INDICATOR_MEMBERSHIP )
    {
//        gsl_integration_workspace * w =
//                gsl_integration_workspace_alloc(1000);

        gsl_integration_cquad_workspace * w =
                gsl_integration_cquad_workspace_alloc(100);


        double result, error;
        double dummyParam = 1.0;
        size_t nEval;

        gsl_function F;
        F.function = &fIntegrate;
        F.params   = this;

        double epsAbs = 1.0e-6;
        double epsRel = 1.0e-6;
        double lb = -1.0;
        double ub = +1.0;

//        std::cout << "Integration sample = " << valueSetSample;

//            gsl_integration_qng(&F,lb,ub,epsAbs,epsRel,&result,&error,&nEval);
//        gsl_integration_qags(&F,lb,ub,epsAbs,epsRel,1000,w,&result,&error);
        gsl_integration_cquad(&F,lb,ub,epsAbs,epsRel,w,&result,&error,&nEval);

//        gsl_integration_workspace_free(w);
        gsl_integration_cquad_workspace_free(w);

//        std::cout << "; result = " << result << "; with error = " << error << std::endl;

        valueMembershipSample = result;

    }

//    std::cout << "valueSetSample = " << valueSetSample << std::endl;
//    std::cout << "valueMembershipSample = " << valueMembershipSample << std::endl;

    // Return output value
    return valueMembershipSample;
}

// Function to be integrated in gsl
double samplingProbabilityConstraint::fIntegrate(double x, void * params)
{
    samplingProbabilityConstraint* classPtr = (samplingProbabilityConstraint*)params;
    double res = classPtr->m_fIntegrate(x);
    return res;
}