/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#ifndef IPANEMA_SAMPLINGPROBABILITYCONSTRAINT_H
#define IPANEMA_SAMPLINGPROBABILITYCONSTRAINT_H

#include <functional>
#include <iostream>

#include "samplingConstraints.h"
#include <gsl/gsl_integration.h>

namespace ipanema
{
namespace ipanemaConstraints
{


    class samplingProbabilityConstraint : public samplingConstraints {

        /**
         * Set membership function enum
         */

    public:

        typedef enum
        {
            INDICATOR_MEMBERSHIP,
            LINEAR_MEMBERSHIP,
            TRIGONOMETRIC_MEMBERSHIP,
            QUADRATIC_MEMBERSHIP,
            BIQUADRATIC_MEMBERSHIP,
            TRIQUADRATIC_MEMBERSHIP,

            MAX_MEMBERSHIP
        } MembershipType;


        /**
         * Parameter struct
         */

    public:

        struct samplingProbabilityConstraintParams : public samplingConstraintsParams {

            // Flag whether expectation functions shall be computed by MC
            bool                                              expectationMC;

            // Probability density values associated to pSamples [if expectationMC = false]
            std::vector<double>*                              pSamplesDensities;

            // Lower bound of set
            double                                            setValueLB;

            // Upper bound of set
            double                                            setValueUB;

            // Function to evaluate whether sample is in arrival zone [setLB<f<setUB] or sample is in outside
            std::function<double(std::vector<double>)>        pSetCondition;

            // Type of membership function to be used
            MembershipType                                    membershipType;

        };



        /**
         * Class members
         */

    protected:

        // Parameters
        const samplingProbabilityConstraintParams* m_pParams;

        // Std pointer to function to integrate
        std::function<double(double)>       m_fIntegrate;


        /**
         * Constructor and Destructor
         */

    public:

        // Default constructor
        samplingProbabilityConstraint(const samplingConstraintsParams* pParams);



        /**
         * List of virtual functions to be implemented in this derived classes
         */

    public:

        // Evaluate the constraint in pSamples
        double evaluate ();



        /**
         * List of private help functions to better organise the code
         */

    private:

        // Compute probability of pSamples to be in set defined by pSetCondition
        double computeSamplesSetProbability();

        // Compute membership of pSamples to be in set defined by pSetCondition
        double computeSampleSetMembership(const std::vector<double> &sample) ;

        // Function to be integrated in gsl
        static double fIntegrate(double x, void * params);


    }; // class samplingProbabilityConstraint


} // namespace ipanemaConstraints
} // namespace ipanema

#endif //IPANEMA_SAMPLINGPROBABILITYCONSTRAINT_H
