/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include "integralOfControl.h"

using namespace std;
using namespace ipanema;
using namespace ipanemaObjectives;



/**
 * Constructor
 */

// Default constructor
integralOfControl::integralOfControl(const deterministicObjectivesParams* pParams) :
        deterministicObjectives(pParams),
        m_pParams(static_cast<const integralOfControlParams*>(pParams))
{
    // Sanity check on dimension
    if ( (*m_pParams->pControlCoefficients).size() !=
            ((*m_pParams->pTimeGrid).size()-1) )
        ipanema_throw("Control coefficients outermost-size and timeGrid size do not match");
}



/**
 * List of virtual functions to implemented in this derived class
 */

// Evaluate the objective according to quantities in params
double integralOfControl::evaluate ()
{
    // Output variable
    double objectiveValue = 0.0;

    // Number of discretisation intervals
    unsigned int ni = (*m_pParams->pTimeGrid).size()-1;

    // Integrate exactly [only for constant control for now]
    double dt;
    for (unsigned int i = 0 ; i < ni; i++)
    {
        dt = (*m_pParams->pTimeGrid)[i+1]-(*m_pParams->pTimeGrid)[i];
        objectiveValue += evaluateIntegrand(i) * dt;
    }

    // Return objective value
    return objectiveValue;
}


/**
 * List of private help functions to better organise the code
 */

// Integrand of integral at discretization interval discretisationIndex
double integralOfControl::evaluateIntegrand (unsigned int index) const
{
    // Output value of integrand
    double integrandValue;

    // Number of control variables
    unsigned int nc = (*m_pParams->pControlCoefficients).size();

    // Control modulus
    double controlModulus = 0.0;
    switch (m_pParams->controlType)
    {
        case COMPONENT_CONTROL:
            for (unsigned int i = 0 ; i < nc; i++)
            {
                controlModulus += pow((*m_pParams->pControlCoefficients)[index][i],2.0);
            }
            controlModulus = sqrt(controlModulus);
            break;
        case MAGNITUDEANGLES_CONTROL:
            controlModulus = abs((*m_pParams->pControlCoefficients)[index][0]);
            break;
        default:
            ipanema_throw("controlType ill-defined");
            break;
    }

    // Integrand
    switch (m_pParams->objectiveIntegrandType)
    {
        case MODULUS_OBJECTIVE:
            integrandValue = abs(controlModulus);
            break;
        case SQUARE_OBJECTIVE:
            integrandValue = pow(controlModulus,2.0);
            break;
        default:
            ipanema_throw("objectiveIntegrandType ill-defined");
            break;
    }

    // Return output value
    return integrandValue;
}