/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#include "ipanemaSolver.h"

using namespace std;
using namespace ipanema;
using namespace ipanemaSolvers;


/**
 * Constructor and Destructor
 */

// Default constructor
ipanemaSolver::ipanemaSolver(ipanemaSolverParams* pParams)
        : m_pParams(pParams), m_g({}), m_x({}), m_J(0), m_gRelax({})
{
    // Sanity check
    if ( (*m_pParams->pControlCoefficients).size() !=
            ((*m_pParams->pTimeGrid).size()-1) )
        ipanema_throw("pControlCoefficients outermost-size and pTimeGrid size not consistent");
    if ( (*m_pParams->pInitialSamples).empty() )
        ipanema_throw("Empty set of initial samples");

    // Set constraint scales if empty otherwise check size
    if (m_pParams->constraintScale.empty())
    {
        m_pParams->constraintScale = std::vector<double>(get_ng(),1.0);
    }
    else if (m_pParams->constraintScale.size()!=get_ng())
    {
        ipanema_throw("Constraint scale has wrong dimension");
    }

    // Set control scales if empty otherwise check size
    if (m_pParams->controlScale.empty())
    {
        m_pParams->controlScale = std::vector<double>(get_nc(),1.0);
    }
    else if (m_pParams->controlScale.size()!=get_nc())
    {
        ipanema_throw("Control scale has wrong dimension");
    }
}

// Default destructor
ipanemaSolver::~ipanemaSolver()
{
    // Nothing here
}


/**
 * List of solver functions
 */

// Only if worhp is enabled
#ifdef __USE_WORHP

int ipanemaSolver::solveByWorhp()
{
    // Make guess feasible
    if (m_pParams->onlyFeasible || m_pParams->optimiseFromFeasible)
    {
        // Make guess feasible
        makeGuessFeasibleByWorhp();

        // If only feasibility, interrupt
        if (m_pParams->onlyFeasible)
            return 0;
    }

    // Print
    cout << endl << "%%%%%%%%%%%%%%%%%%%%%%%%%%%% Optimisation with Objective and Constraints %%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl << endl;

    // Dimension of problem
    unsigned int nSamples = (*m_pParams->pInitialSamples).size();   // Number of samples
    unsigned int ns = (*m_pParams->pInitialSamples)[0].size();      // Number of state variables
    unsigned int nc = get_nc();                                     // Number of control variables
    unsigned int ni = get_ni();                                     // Number of discretisation intervals
    unsigned int ng = get_ng();                                     // Number of constraints
    unsigned int nx = get_nx();                                     // Number of free variables

    // Clear output variables
    m_J = 0.0;
    m_g.clear();
    m_x.clear();
    m_gRelax = std::vector<double>(ng,0.0);
    

    /**
     * WORHP CALL
     */

    int status;

    /* Worhp data structures */
    OptVar opt;
    Workspace wsp;
    Params par;
    Control cnt;

    /* Check version of library and header files */
    CHECK_WORHP_VERSION

    /*
     * Properly zeros everything, or else the following routines
     * could get confused.
     */
    WorhpPreInit(&opt, &wsp, &par, &cnt);
    
    /* Make sure so the initialisers know they have to work! */
    opt.initialised = false;
    wsp.initialised = false;
    par.initialised = false;
    cnt.initialised = false;

    /* Uncomment this to get more info on data structures */
    /*  WorhpDiag(&opt, &wsp, &par, &cnt); */

    /*
     * Parameter initialisation routine that must be called
     * when using ReadParamsNoInit instead of ReadParams->
     */
    InitParams(&status, &par);

    /*
     * We can now set parameters that may be overruled by those in the
     * parameter file. This is useful for setting a non-default standard
     * parameter value that may still be overwritten.
     */
    
    /*
     * Parameter XML import routine that does not reset
     * all parameters to default values (InitParams does this)
     */
    ReadParamsNoInit(&status, "worhp.xml", &par);

    if (status == DataError || status == InitError)
        return EXIT_FAILURE;

    /*
     * We can now set parameters that may be overruled by those in the
     * parameter file. This is useful for setting a non-default standard
     * parameter value that may still be overwritten.
     */
//    par.NLPprint = set.print;

    /* Specify sparse matrix format */
    par.MatrixCC = false;

    status = OK; /* ignore other errors: falls back to standard values. */

    /* Overload user derivative input flags */
    par.UserDF = false;
    par.UserDG = false;
    par.UserHM = false;

    par.CheckValuesDF = false;
    par.CheckStructureDF = true;

    par.CheckValuesDG = false;
    par.CheckStructureDG = true;

    par.CheckValuesHM = false;
    par.CheckStructureHM = true;

    /* Overload custom for feasibility only */
    par.FeasibleOnly = false;

//            par.CheckDerivIter = set.checkIter;

    /* Specify number of variables and constraints. */
    opt.n = nx;
    opt.m = ng;
    
    /* Specify nonzeros of derivative matrices. Accordingly to the manual
       independently to the chosen format. */
    // Full dependency
    wsp.DF.nnz = nx;
    wsp.DG.nnz = ng*nx;
    if (par.UserHM || par.FidifHM || par.BFGSmethod == 100)
        wsp.HM.nnz = (nx*(nx+1))/2; // Lower-triangular pattern

    /* Data structure initialisation. */
    WorhpInit(&opt, &wsp, &par, &cnt);
    if (cnt.status != FirstCall) {
        printf("%s\n", "Main: Initialisation failed.");
        return EXIT_FAILURE;
    }


    /* Initial estimates for X and Lambda. Define bounds for X */
    std::vector<double> initial_x = get_xFromGuess();
    std::vector<double> xLB, xUB;
    get_xBounds(xLB,xUB);

    for (int i = 0; i < opt.n; i++) {
        opt.X[i] = initial_x[i];
        opt.Lambda[i] = 0.0;

        opt.XL[i] = xLB[i];
        opt.XU[i] = xUB[i];

        printf("%+5.5f %+5.5f %+5.5f",opt.X[i],opt.XL[i]-opt.X[i],opt.X[i]-opt.XU[i]);
    }


    // Clear
    initial_x.clear();
    xLB.clear();
    xUB.clear();


    /* Define bounds for G and estimates for Mu */
    std::vector<double> gLB, gUB;
    get_gBounds(gLB,gUB);
    for (int i = 0; i < opt.m; i++) {
        opt.Mu[i] = 0.0;

        if (gLB[i] < -1.0e20)
            opt.GL[i] = -par.Infty;
        else
            opt.GL[i] = gLB[i];

        if (gUB[i] > +1.0e20)
            opt.GU[i] = +par.Infty;
        else
            opt.GU[i] = gUB[i];
    }
    gLB.clear();
    gUB.clear();


    /*
     * Specify matrix structures in CS or CC format depending on input, using Fortran indexing,
     * i.e. 1...N instead of 0...N-1, to describe the matrix structure.
     * (This is more elegant in Fortran.)
     */

    /* Gradient of f */
    // Full vector
    if (wsp.DF.NeedStructure) {
        for (int i = 0; i < wsp.DF.nnz; i++)
            wsp.DF.row[i] = i+1;
        std::cout << "Structure of DF set" << std::endl;
    }

    /* Gradient of g */
    // Full matrix
    if (wsp.DG.NeedStructure) {
        for (int i = 0, cg = 0; i < nx; i++)
        {
            for (int j = 0; j < ng; j++)
            {
                wsp.DG.row[cg]   = j+1;
                wsp.DG.col[cg++] = i+1;
            }
        }
        std::cout << "Structure of DG set" << std::endl;
    }

    /* Hessian */
    // Full lower-diagonal matrix
    if (wsp.HM.NeedStructure) {
        int cH = 0;
        for (int i = 0; i < nx-1; i++)
        {
            for (int j = i+1; j < nx; j++)
            {
                wsp.HM.row[cH]   = j+1;
                wsp.HM.col[cH++] = i+1;
            }
        }
        for (int i = 0 ; i < nx; i++)
        {
            wsp.HM.row[cH]   = i+1;
            wsp.HM.col[cH++] = i+1;
        }
        std::cout << "Structure of HM set" << std::endl;
    }

    /* Uncomment this to have a look at the workspace slices */
    /*  PrintIWMT(wsp); */
    /*  PrintRWMT(wsp); */

    // Worhp cycle
    std::vector<double> x(nx,0.0); // Free variable vector
    unsigned int cIter = 0;
    while (cnt.status < TerminateSuccess && cnt.status > TerminateError) {

        if (GetUserAction(&cnt, callWorhp)) {
            Worhp(&opt, &wsp, &par, &cnt);
            /* No DoneUserAction! */
        }

        if (GetUserAction(&cnt, iterOutput)) {
            IterationOutput(&opt, &wsp, &par, &cnt);
            DoneUserAction(&cnt, iterOutput);
        }

        /**
         * Transcribe
         */

        //// Check if new free variable is different as in Major iteration
        //
        // Start with major iteration to false
        (*m_pParams->pMajorIteration) = false;
        for (int i = 0 ; i < opt.n && cIter>0 ; i++)
        {
            if (std::abs(opt.X[i] - x[i]) > (par.FidifEps + 1.0e-6))
            {
                (*m_pParams->pMajorIteration) = true;
            }
        }
        
        // Update free variables
        for (int i = 0; i < opt.n; i++)
            x[i] = opt.X[i];

//        std::cout << "Free variable x = " ;
//        for (int i = 0; i < opt.n; i++)
//            std::cout << x[i] << " ";
//        std::cout << std::endl;

        // Transcribe to obj and constraint
        transcribe(x);
        


        /**
         * Update requested quantities
         */

        if (GetUserAction(&cnt, evalF)) {

            // Update objective value
            opt.F = wsp.ScaleObj * get_J();
            DoneUserAction(&cnt, evalF);
        }

        if (GetUserAction(&cnt, evalG)) {

            // Update constraint value
            std::vector<double> g = get_g();
            for ( int i = 0 ; i < opt.m ; i++ ) {
                opt.G[i] = g[i];
            }
            DoneUserAction(&cnt, evalG);
        }

        if (GetUserAction(&cnt, evalDF)) {

            ipanema_throw("Shall not enter DF");
            DoneUserAction(&cnt, evalDF);
        }

        if (GetUserAction(&cnt, evalDG)) {

            ipanema_throw("Shall not enter DG");
            DoneUserAction(&cnt, evalDG);
        }

        if (GetUserAction(&cnt, evalHM)) {
            ipanema_throw("Shall not enter HM");
            DoneUserAction(&cnt, evalHM);
        }

        if (GetUserAction(&cnt, fidif)) {
            WorhpFidif(&opt, &wsp, &par, &cnt);
            /* No DoneUserAction! */
        }

//        cout << endl << "Evaluation for x = " << x << endl;
//        cout << "F = " << opt.F << endl;
//        cout << "G = " << get_g() << endl;
//        cout << "DF = " ;
//        for ( auto i = 0 ; i < nx; i++ )
//            cout << wsp.DF.val[i] << " " ;
//        cout << endl;
//        cout << "DG = " ;
//        for ( auto i = 0 ; i < ng*nx; i++ )
//            cout << wsp.DG.val[i] << " " ;
//        cout << endl << endl;

        // Update iterator count
        cIter++;
    }

    int worhp_status = cnt.status;

    StatusMsg(&opt, &wsp, &par, &cnt);
    WorhpFree(&opt, &wsp, &par, &cnt);

    return worhp_status;

}


int ipanemaSolver::makeGuessFeasibleByWorhp()
{
    // Print
    cout << endl << "%%%%%%%%%%%%%%%%%%%%%%%%%%%% Optimisation in Feasibility Only Mode %%%%%%%%%%%%%%%%%%%%%%%%%%%%" << endl << endl;

    // Dimension of problem
    unsigned int nSamples = (*m_pParams->pInitialSamples).size();   // Number of samples
    unsigned int ns = (*m_pParams->pInitialSamples)[0].size();      // Number of state variables
    unsigned int nc = get_nc();                                     // Number of control variables
    unsigned int ni = get_ni();                                     // Number of discretisation intervals
    unsigned int ng = get_ng();                                     // Number of constraints
    unsigned int nx = get_nx();                                     // Number of free variables

    // Clear output variables
    m_J = 0.0;
    m_g.clear();
    m_x.clear();
    m_gRelax = std::vector<double>(ng,0.0);


    /**
     * WORHP CALL
     */

    int status;

    /* Worhp data structures */
    OptVar opt;
    Workspace wsp;
    Params par;
    Control cnt;

    /* Check version of library and header files */
    CHECK_WORHP_VERSION

    /*
     * Properly zeros everything, or else the following routines
     * could get confused.
     */
    WorhpPreInit(&opt, &wsp, &par, &cnt);

    /* Make sure so the initialisers know they have to work! */
    opt.initialised = false;
    wsp.initialised = false;
    par.initialised = false;
    cnt.initialised = false;

    /* Uncomment this to get more info on data structures */
    /*  WorhpDiag(&opt, &wsp, &par, &cnt); */

    /*
     * Parameter initialisation routine that must be called
     * when using ReadParamsNoInit instead of ReadParams->
     */
    InitParams(&status, &par);

    /*
     * We can now set parameters that may be overruled by those in the
     * parameter file. This is useful for setting a non-default standard
     * parameter value that may still be overwritten.
     */

    /*
     * Parameter XML import routine that does not reset
     * all parameters to default values (InitParams does this)
     */
    ReadParamsNoInit(&status, "worhp.xml", &par);

    if (status == DataError || status == InitError)
        return EXIT_FAILURE;

    /*
     * We can now set parameters that may be overruled by those in the
     * parameter file. This is useful for setting a non-default standard
     * parameter value that may still be overwritten.
     */
//    par.NLPprint = set.print;

    /* Specify sparse matrix format */
    par.MatrixCC = false;

    status = OK; /* ignore other errors: falls back to standard values. */

    /* Overload user derivative input flags */
    par.UserDF = false;
    par.UserDG = false;
    par.UserHM = false;

    par.CheckValuesDF = false;
    par.CheckStructureDF = true;

    par.CheckValuesDG = false;
    par.CheckStructureDG = true;

    par.CheckValuesHM = false;
    par.CheckStructureHM = true;

    /* Overload feasibility only */
    par.FeasibleOnly = true;

//            par.CheckDerivIter = set.checkIter;

    /* Specify number of variables and constraints. */
    opt.n = nx;
    opt.m = ng;

    /* Specify nonzeros of derivative matrices. Accordingly to the manual
       independently to the chosen format. */
    // Full dependency
    wsp.DF.nnz = nx;
    wsp.DG.nnz = ng*nx;
    if (par.UserHM || par.FidifHM || par.BFGSmethod == 100)
        wsp.HM.nnz = (nx*(nx+1))/2; // Lower-triangular pattern

    /* Data structure initialisation. */
    WorhpInit(&opt, &wsp, &par, &cnt);
    if (cnt.status != FirstCall) {
        printf("%s\n", "Main: Initialisation failed.");
        return EXIT_FAILURE;
    }


    /* Initial estimates for X and Lambda. Define bounds for X */
    std::vector<double> initial_x = get_xFromGuess();
    std::vector<double> xLB, xUB;
    get_xBounds(xLB,xUB);

    for (int i = 0; i < opt.n; i++) {
        opt.X[i] = initial_x[i];
        opt.Lambda[i] = 0.0;

        opt.XL[i] = xLB[i];
        opt.XU[i] = xUB[i];
//        printf("%+5.5f %+5.5f %+5.5f\n",opt.X[i],opt.XL[i]-opt.X[i],opt.X[i]-opt.XU[i]);
    }
    initial_x.clear();
    xLB.clear();
    xUB.clear();


    /* Define bounds for G and estimates for Mu */
    std::vector<double> gLB, gUB;
    get_gBounds(gLB,gUB);
    for (int i = 0; i < opt.m; i++) {
        opt.Mu[i] = 0.0;

        if (gLB[i] < -1.0e20)
            opt.GL[i] = -par.Infty;
        else
            opt.GL[i] = gLB[i];

        if (gUB[i] > +1.0e20)
            opt.GU[i] = +par.Infty;
        else
            opt.GU[i] = gUB[i];
    }
    gLB.clear();
    gUB.clear();


    /*
     * Specify matrix structures in CS or CC format depending on input, using Fortran indexing,
     * i.e. 1...N instead of 0...N-1, to describe the matrix structure.
     * (This is more elegant in Fortran.)
     */

    /* Gradient of f */
    // Full vector
    if (wsp.DF.NeedStructure) {
        for (int i = 0; i < wsp.DF.nnz; i++)
            wsp.DF.row[i] = i+1;
        std::cout << "Structure of DF set" << std::endl;
    }

    /* Gradient of g */
    // Full matrix
    if (wsp.DG.NeedStructure) {
        for (int i = 0, cg = 0; i < nx; i++)
        {
            for (int j = 0; j < ng; j++)
            {
                wsp.DG.row[cg]   = j+1;
                wsp.DG.col[cg++] = i+1;
            }
        }
        std::cout << "Structure of DG set" << std::endl;
    }

    /* Hessian */
    // Full lower-diagonal matrix
    if (wsp.HM.NeedStructure) {
        int cH = 0;
        for (int i = 0; i < nx-1; i++)
        {
            for (int j = i+1; j < nx; j++)
            {
                wsp.HM.row[cH]   = j+1;
                wsp.HM.col[cH++] = i+1;
            }
        }
        for (int i = 0 ; i < nx; i++)
        {
            wsp.HM.row[cH]   = i+1;
            wsp.HM.col[cH++] = i+1;
        }
        std::cout << "Structure of HM set" << std::endl;
    }

    /* Uncomment this to have a look at the workspace slices */
    /*  PrintIWMT(wsp); */
    /*  PrintRWMT(wsp); */

    // Worhp cycle
    std::vector<double> x(nx,0.0); // Free variable vector
    unsigned int cIter = 0;
    while (cnt.status < TerminateSuccess && cnt.status > TerminateError) {

        if (GetUserAction(&cnt, callWorhp)) {
            Worhp(&opt, &wsp, &par, &cnt);
            /* No DoneUserAction! */
        }

        if (GetUserAction(&cnt, iterOutput)) {
            IterationOutput(&opt, &wsp, &par, &cnt);
            DoneUserAction(&cnt, iterOutput);
        }

        /**
         * Transcribe
         */

        //// Check if new free variable is different as in Major iteration
        //
        // Start with major iteration to false
        (*m_pParams->pMajorIteration) = false;
        for (int i = 0 ; i < opt.n && cIter>0 ; i++)
        {
            if (std::abs(opt.X[i] - x[i]) > (par.FidifEps + 1.0e-6))
            {
                (*m_pParams->pMajorIteration) = true;
            }
        }

        // Update free variables
        for (int i = 0; i < opt.n; i++)
            x[i] = opt.X[i];

//        std::cout << "Free variable x = " ;
//        for (int i = 0; i < opt.n; i++)
//            std::cout << x[i] << " ";
//        std::cout << std::endl;

        // Transcribe to obj and constraint
        transcribe(x);



        /**
         * Update requested quantities
         */

        if (GetUserAction(&cnt, evalF)) {

            // Update objective value
            opt.F = wsp.ScaleObj * get_J();
            DoneUserAction(&cnt, evalF);
        }

        if (GetUserAction(&cnt, evalG)) {

            // Update constraint value
            std::vector<double> g = get_g();
            for ( int i = 0 ; i < opt.m ; i++ ) {
                opt.G[i] = g[i];
            }
            DoneUserAction(&cnt, evalG);
        }

        if (GetUserAction(&cnt, evalDF)) {

            ipanema_throw("Shall not enter DF");
            DoneUserAction(&cnt, evalDF);
        }

        if (GetUserAction(&cnt, evalDG)) {

            ipanema_throw("Shall not enter DG");
            DoneUserAction(&cnt, evalDG);
        }

        if (GetUserAction(&cnt, evalHM)) {
            ipanema_throw("Shall not enter HM");
            DoneUserAction(&cnt, evalHM);
        }

        if (GetUserAction(&cnt, fidif)) {
            WorhpFidif(&opt, &wsp, &par, &cnt);
            /* No DoneUserAction! */
        }

//        cout << endl << "Evaluation for x = " << x << endl;
//        cout << "F = " << opt.F << endl;
//        cout << "G = " << get_g() << endl;
//        cout << "DF = " ;
//        for ( auto i = 0 ; i < nx; i++ )
//            cout << wsp.DF.val[i] << " " ;
//        cout << endl;
//        cout << "DG = " ;
//        for ( auto i = 0 ; i < ng*nx; i++ )
//            cout << wsp.DG.val[i] << " " ;
//        cout << endl << endl;

        // Update iterator count
        cIter++;
    }

    int worhp_status = cnt.status;

    StatusMsg(&opt, &wsp, &par, &cnt);
    WorhpFree(&opt, &wsp, &par, &cnt);

    return worhp_status;

}

#endif // __USE_WORHP


// Transcribe continuous quantities into objective function and constraint
int ipanemaSolver::transcribe(const std::vector<double> &x)
{
    // Set x
    m_x = x;

    // Dimension
    unsigned int ng = get_ng();
    
    // Set guess from NLP vector x
    set_guessFromX(x);
    
    // Propagate the samples
    propagateSamples();
    
    // Evaluate objective function and update
    m_J = m_pParams->pObjective() / m_pParams->objectiveScale;
    
    // Penalise constraint relaxation on objective function
    if (m_pParams->constraintRelax)
    {
        for (unsigned int i = 0 ; i < ng; i++)
        {
            m_J += std::abs(m_pParams->gRelaxPenalty[i]*m_gRelax[i]);
        }
    }
    
    // Evaluate constraints and update
    m_g.clear();
    m_g = std::vector<double> (ng);
    for (unsigned int i = 0 ; i < ng; i++)
    {
        m_g[i] = m_pParams->pConstraints[i]() / m_pParams->constraintScale[i];

        // Remove relaxation on constraint
        if (m_pParams->constraintRelax)
            m_g[i] -= m_gRelax[i];
    }
    
    // Return success
    return 0;
}

/**
 * List of private help functions to better organise the code
 */

// Function that propagate samples with pPropagator
void ipanemaSolver::propagateSamples()
{
    // Number of samples to propagate
    unsigned int nSamples = (*m_pParams->pInitialSamples).size();   // Number of samples
    
    // Propagate samples through propagator
    (*m_pParams->pFinalSamples).clear();
    (*m_pParams->pFinalSamples) = (*m_pParams->pInitialSamples);
    for (unsigned int i = 0 ; i < nSamples; i++)
    {
        (*m_pParams->pFinalSamples)[i] =
            m_pParams->pPropagator((*m_pParams->pInitialSamples)[i]);
    }
}


// Get initial free variable vector from guess
std::vector<double> ipanemaSolver::get_xFromGuess() const
{
    // Output vector
    std::vector<double> x;

    // Dimensions
    unsigned int ni = get_ni(); // Get number of discretisation intervals
    unsigned int nc = get_nc(); // Get number of control variables

    for (unsigned int i = 0; i < ni; i++)
    {
        for (unsigned int j = 0 ; j < nc; j++)
        {
            x.push_back(
                    (*m_pParams->pControlCoefficients)[i][j] / m_pParams->controlScale[j]
            );
        }
    }

    // Constraint relaxation is free variable
    if (m_pParams->constraintRelax)
    {
        unsigned int ng = get_ng();
        for (unsigned int i = 0 ; i < ng; i++)
        {
            x.push_back(0.0);
        }
    }

    // Sanity check
    if (x.size()!=get_nx())
        ipanema_throw("Error in initial free variable vector computation");

    // Return output value
    return x;
}

// Get free variables bounds
void ipanemaSolver::get_xBounds(std::vector<double> &xLB, std::vector<double> &xUB) const
{
    // Output vectors
    xLB.clear();
    xUB.clear();

    // Dimensions
    unsigned int ni = get_ni(); // Get number of discretisation intervals
    unsigned int nc = get_nc(); // Get number of control variables

    // Assign output values
    for (unsigned int i = 0; i < ni; i++)
    {
        for (unsigned int j = 0 ; j < nc; j++)
        {
            xLB.push_back(m_pParams->controlBounds[j][0] / m_pParams->controlScale[j]);
            xUB.push_back(m_pParams->controlBounds[j][1] / m_pParams->controlScale[j]);
        }
    }

    // Constraint relaxation is free variable
    if (m_pParams->constraintRelax)
    {
        unsigned int ng = get_ng();
        for (unsigned int i = 0 ; i < ng; i++)
        {
            xLB.push_back(0.0);
            xUB.push_back(m_pParams->gRelaxUB);
        }
    }

    // Sanity check
    if (xLB.size()!=get_nx() || xUB.size()!=get_nx())
        ipanema_throw("Error in free variable bounds computation");
}

// Get constraints bounds
void ipanemaSolver::get_gBounds(std::vector<double> &gLB, std::vector<double> &gUB) const
{
    // Output vectors
    gLB.clear();
    gUB.clear();

    // Dimensions
    unsigned int ng = get_ng(); // Get number of constraints

    // Sanity check
    if (m_pParams->constraintsBounds.size()!=ng)
        ipanema_throw("Number of constraints bounds differ from number of constraints");

    // Assign output values
    for (unsigned int i = 0; i < ng; i++)
    {
        gLB.push_back( m_pParams->constraintsBounds[i][0] / m_pParams->constraintScale[i] );
        gUB.push_back( m_pParams->constraintsBounds[i][1] / m_pParams->constraintScale[i] );
    }

    // Sanity check
    if (gLB.size()!=ng || gUB.size()!=ng)
        ipanema_throw("Error in free variable bounds computation");
}

// Set guess from free variable vector of NLP solver
void ipanemaSolver::set_guessFromX(const std::vector<double> &x)
{
    // Sanity check
    if(x.size()!=get_nx())
        ipanema_throw("Input free variable vector of NLP solver has not required number of elements");

    // Dimensions
    unsigned int ni = get_ni(); // Get number of discretisation intervals
    unsigned int nc = get_nc(); // Get number of control variables

    // Counter
    unsigned int cx = 0;

    // Control parameters
    for (unsigned int i = 0; i < ni; i++)
    {
        for (unsigned int j = 0 ; j < nc; j++)
        {
            (*m_pParams->pControlCoefficients)[i][j] = x[cx++] * m_pParams->controlScale[j];
        }
    }

    // Constraint relaxation is free variable
    if (m_pParams->constraintRelax)
    {
        unsigned int ng = get_ng();
        for (unsigned int i = 0 ; i < ng; i++)
        {
            m_gRelax[i] = x[cx++];
        }
    }

}

// Get number of free variables [for now only constant control]
unsigned int ipanemaSolver::get_nx () const
{
    // Output value
    unsigned int nx = 0;

    // Dimension of problem
    unsigned int nc = (*m_pParams->pControlCoefficients)[0].size(); // Number of control variables
    unsigned int ni = (*m_pParams->pTimeGrid).size()-1;             // Number of discretisation intervals

    // Control parameters
    nx = nc*ni;

    // Relaxation values
    if (m_pParams->constraintRelax)
    {
        nx += get_ng();
    }

    // Return output value
    return nx;
}

// Get number of constraints
unsigned int ipanemaSolver::get_ng () const
{
    // Return output value [constraint are scalar]
    return (m_pParams->pConstraints.size());
}

// Get number of controls
unsigned int ipanemaSolver::get_nc () const
{
    return ((*m_pParams->pControlCoefficients)[0].size());
}

// Get number of discretisation intervals
unsigned int ipanemaSolver::get_ni () const
{
    return ((*m_pParams->pTimeGrid).size()-1);
}


/**
 * Functions to get output variables
 */

// Get free variable vector
std::vector<double> ipanemaSolver::get_x() const
{
    return m_x;
}

// Get objective
double ipanemaSolver::get_J() const
{
    return m_J;
}

// Get constraints vector
std::vector<double> ipanemaSolver::get_g() const
{
    return m_g;
}

// Get constraints relaxation
std::vector<double> ipanemaSolver::get_gRelax() const
{
    return m_gRelax;
}