/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
/*
------ Copyright (C) 2018 University of Strathclyde and Authors ------
-------------------- e-mail: c.greco@strath.ac.uk --------------------
----------------------- Author: Cristian Greco -----------------------
*/

#ifndef IPANEMA_IPANEMASOLVER_H
#define IPANEMA_IPANEMASOLVER_H

#include <vector>
#include <functional>
#include <iostream>

#include "../ipanemaException.h"
#include "Utils/mixed_functions.h"

#ifdef __USE_WORHP
    #include <worhp/worhp.h>
#endif

namespace ipanema
{
namespace ipanemaSolvers
{

    class ipanemaSolver {

        /**
         * Parameter struct
         */

    public:

        struct ipanemaSolverParams {

            // Control parematers [constant control]
            std::vector<std::vector<double>>*           pControlCoefficients;

            // Time discretisation (or any other independent variable) [same as propagator]
            std::vector<double>*                        pTimeGrid;

            // Set of initial samples
            std::vector<std::vector<double>>*           pInitialSamples;

            // Set of final samples [same as sampling constraints]
            std::vector<std::vector<double>>*           pFinalSamples;

            // Function pointer to propagate initial samples to final time
            std::function<std::vector<
                    double>(std::vector<double>)>       pPropagator;

            // Function to evaluate objective
            std::function<double()>                     pObjective;

            // Vector of function pointers to evaluate constraints
            std::vector<std::function<double()>>        pConstraints;

            // Constraints bounds
            std::vector<std::vector<double>>            constraintsBounds;

            // Control bounds
            std::vector<std::vector<double>>            controlBounds;

            // Control scale
            std::vector<double>                         controlScale = {};

            // Constraint scale
            std::vector<double>                         constraintScale = {};

            // Objective scale
            double                                      objectiveScale = 1.0;

            // NLP Major iteration bool
            bool*                                       pMajorIteration;

            // Bool whether to enable constraints relaxation
            bool                                        constraintRelax;

            // Constraint relaxation penalty factors on Objective Function
            std::vector<double>                         gRelaxPenalty;

            // Constraint relax upper bound [lower=0.0]
            double                                      gRelaxUB;

            // Feasibility first then optimise
            bool                                        optimiseFromFeasible = false;

            // Only feasibility
            bool                                        onlyFeasible = false;

        };



        /**
         * Class members
         */

    protected:

        // Parameters
        ipanemaSolverParams*  m_pParams;

        // Current free variable vector
        std::vector<double>         m_x;

        // Current objective value
        double                      m_J;

        // Current constraints value
        std::vector<double>         m_g;

        // Current constraints relaxation value [if constraintRelax = true]
        std::vector<double>         m_gRelax;



        /**
         * Constructor and Destructor
         */

    public:

        // Default constructor
        ipanemaSolver(ipanemaSolverParams* pParams);

        // Default destructor
        virtual ~ipanemaSolver();


        /**
         * List of solver functions
         */

    public:

        // Only if worhp is enabled
        int solveByWorhp();

        // Only if worhp is enabled
        int makeGuessFeasibleByWorhp();


        /**
         * List of private help functions to better organise the code
         */

    private:

        // Transcribe continuous quantities into objective function and constraint
        int transcribe(const std::vector<double> &x);

        // Function that propagate samples with pPropagator
        void propagateSamples() ;

        // Get initial free variable vector from guess
        std::vector<double> get_xFromGuess() const;

        // Get free variables bounds
        void get_xBounds(std::vector<double> &xLB, std::vector<double> &xUB) const;

        // Get constraints bounds
        void get_gBounds(std::vector<double> &gLB, std::vector<double> &gUB) const;

        // Set guess from free variable vector of NLP solver
        void set_guessFromX(const std::vector<double> &x);

        // Get number of free variables
        unsigned int get_nx () const;

        // Get number of constraints
        unsigned int get_ng () const;

        // Get number of controls
        unsigned int get_nc () const;

        // Get number of discretisation intervals
        unsigned int get_ni () const;



        /**
         * Functions to get output variables
         */

    public:

        // Get free variable vector
        std::vector<double> get_x() const;

        // Get objective
        double get_J() const;

        // Get constraints vector
        std::vector<double> get_g() const;

        // Get constraints relaxation
        std::vector<double> get_gRelax() const;

    };

}
}


#endif //IPANEMA_IPANEMASOLVER_H
